#
# (C) Copyright 2018 InovaDevelopment.comn
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl Schopmeyer <inovadevelopment.com>
#

"""
Mock support for the WBEMConnection class that adds the methods that
process a class by incorporating any inherited qualifiers, properties, methods,
and parameters from the superclass so that the class contains all of the
properties, methods, parameters, and qualifiers that reflect this inheritance
and has the class_origin attribute and the propagated attribute correctly set
in the resulting resolved class.

This resolves the combination of the class ancestry with the definition of the
new class to create a what the user expects in the results from get_class, etc.

NOTE: The repository methods will determine when a class is resolved, as it
is inserted into the repository by CreateClass or as it is removed with the
GetClass or EnumerateClass methods.

For documentation, see mocksupport.rst.
"""

from __future__ import absolute_import, print_function

import six

from pywbem import CIMError, CIM_ERR_INVALID_PARAMETER, CIM_ERR_NOT_FOUND, \
    CIM_ERR_INVALID_SUPERCLASS, CIMParameter, CIMMethod, CIMProperty

from pywbem._utils import _format

from pywbem._nocasedict import NocaseDict


class ResolverMixin(object):  # pylint: disable=too-few-public-methods
    """
    Mixin class that adds the methods for resolving class and the properties,
    methods, and parameters that are contained in the classes to be useful
    in the repository.

    As classes are passed to the create_class method or through the compiler
    into the mock repository they are incomplete (they define the properties
    and methods of the defined class as well as a possible superclass
    definition, and qualifiers (ex. override) that define relationships with
    superclasses). When a user requests or other methods like getinstance uses
    a class from the repository they depend on all of the properties, methods,
    parameters, and qualifiers to be properly resolved so that the class is
    viewed in total. This requires further processing (resolution) to fit into
    the repository including:

    1. Inheritance of properties and methods from superclasses.

    2. Validation that the class is legally defined and that the dependences
       exist.

    3. Consolidation of superclass definitions (properties, methods, qualifiers)
       into the class.

    4. Setting of the characteristics of property, method, and qualifier
       attributes such as class_origin, and propagated properties.

    """
    @staticmethod
    def _test_qualifier_decl(qualifier, qualifier_repo, namespace):
        """
        Test that qualifier is in repo and valid.
        For for conn_lite, ignore this test
        """
        if qualifier_repo is None:
            return
        if qualifier.name not in qualifier_repo:
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format("Qualifier declaration {0!A} required by CreateClass "
                        "not found in namespace {1!A}.",
                        qualifier.name, namespace))

    @staticmethod
    def _validate_qualifiers(qualifier_list, qual_repo, new_class, scope):
        """
        Validate a list of qualifiers against the Qualifier decl in the
        repository.

        1. Whether it is declared (can be obtained from the declContext).
        2. Whether it has the same type as the declaration.
        3. Whether the qualifier is valid for the given scope.
        4. Whether the qualifier can be overridden.
        5. Whether the qualifier should be propagated to the subclass.
        """
        for qname, qvalue in qualifier_list.items():
            if qname not in qual_repo:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("Qualifier {0|A} in new_class {1|A} in "
                            "CreateClass not in repository.",
                            (qname, new_class.classnam)))
            q_decl = qual_repo[qname]
            if qvalue.type != q_decl.type:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("Qualifier {0|A} in new_class {1|A} override type "
                            "mismatch {2|A} with qualifier declaration {3|A}.",
                            (qname, new_class.classname, qvalue.type,
                             q_decl.type)))
            if scope not in q_decl.scopes:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("Qualifier {0|A} in new class {1|A} scope {2|A} "
                            "invalid. Not in qualifier decl scopes {3}",
                            (qname, new_class.classname, scope, q_decl.scopes)))

    @staticmethod
    def _init_qualifier(qualifier, qual_repo):
        """
        Initialize the flavors of a qualifier from the qualifier repo and
        initialize propagated.
        """
        qual_dict_entry = qual_repo[qualifier.name]
        qualifier.propagated = False
        if qualifier.tosubclass is None:
            if qual_dict_entry.tosubclass is None:
                qualifier.tosubclass = True
            else:
                qualifier.tosubclass = qual_dict_entry.tosubclass
        if qualifier.overridable is None:
            if qual_dict_entry.overridable is None:
                qualifier.overridable = True
            else:
                qualifier.overridable = qual_dict_entry.overridable
        if qualifier.translatable is None:
            qualifier.translatable = qual_dict_entry.translatable

    @staticmethod
    def _init_qualifier_decl(qualifier_decl, qual_repo):
        """
        Initialize the flavors of a qualifier declaration if they are not
        already set.
        """
        assert qualifier_decl.name not in qual_repo
        if qualifier_decl.tosubclass is None:
            qualifier_decl.tosubclass = True
        if qualifier_decl.overridable is None:
            qualifier_decl.overridable = True
        if qualifier_decl.translatable is None:
            qualifier_decl.translatable = False

    def _resolve_objects(self, new_objects, superclass_objects, new_class,
                         superclass, classrepo, qualifier_repo, type_str,
                         verbose=None):
        """
        Resolve a dictionary of objects where the objects can be CIMProperty,
        CIMMethod, or CIMParameter.  This method resolves each of the objects
        in the dictionary, using the superclass if it is defined.
        """
        if new_objects:   # TODO Future REMOVE. This is test code
            assert isinstance(new_objects, (dict, NocaseDict))
            keys = new_objects.keys()
            assert isinstance(new_objects[keys[0]], (CIMMethod, CIMProperty,
                                                     CIMParameter))

        if not superclass:
            for new_obj in six.itervalues(new_objects):
                self._set_new_object(new_obj, None, new_class, None,
                                     qualifier_repo,
                                     False, type_str)
            return

        # process objects if superclass exists
        for oname, new_obj in six.iteritems(new_objects):
            if oname not in superclass_objects:
                self._set_new_object(new_obj, None, new_class,
                                     superclass, qualifier_repo,
                                     False, type_str)
                continue

            # oname in superclass_objects
            # TODO: We may have object naming because of override.
            if 'Override' not in new_objects[oname].qualifiers:
                if not isinstance(new_objects[oname], CIMParameter):
                    raise CIMError(
                        CIM_ERR_INVALID_PARAMETER,
                        _format("{0} {1!A} in {2!A} duplicates {0} in "
                                "{3!A} without override.",
                                type_str, oname, new_class.classname,
                                superclass.classname))

                # TODO need to finish this.  For now just let
                # parameter slide. Keep the new one.
                continue

            # process object override
            # get override name
            override_name = new_objects[oname].qualifiers["override"].value
            if isinstance(new_obj, (CIMParameter, CIMProperty)):
                if new_obj.type == 'reference':
                    if override_name != oname:
                        raise CIMError(
                            CIM_ERR_INVALID_PARAMETER,
                            _format("Invalid new_class reference "
                                    "{0} {1!A}. in class {2!A}"
                                    "Override must not change {0} "
                                    "name but override name is {3!A}",
                                    type_str, oname, superclass.classname,
                                    override_name))
            try:
                super_obj = superclass_objects[override_name]
            except KeyError:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("Invalid new_class override  {0} {1!A}. in class "
                            "{2!A}. Override name {3!A}} not found in  {3!A}.",
                            type_str, oname, new_class.classname, override_name,
                            superclass.classname))

            # Test if new object characteristics consistent with
            # requirements for that object type
            if isinstance(super_obj, CIMProperty):
                if super_obj.type != new_obj.type \
                    or super_obj.is_array != new_obj.is_array \
                        or super_obj.embedded_object != \
                        new_obj.embedded_object:
                    raise CIMError(
                        CIM_ERR_INVALID_PARAMETER,
                        _format("Invalid new_class property {0!A}. "
                                "Does not match overridden property "
                                "{1!A} in class {2!A}",
                                oname, super_obj.name,
                                superclass.classname))
            elif isinstance(super_obj, CIMMethod):
                if super_obj.return_type != new_obj.return_type:
                    raise CIMError(
                        CIM_ERR_INVALID_PARAMETER,
                        _format("Invalid new_class method {0!A}. "
                                "Mismatch method  return typein "
                                "class {1!A}.", oname,
                                superclass.classname))
            elif isinstance(super_obj, CIMParameter):
                if super_obj.type != new_obj.type or \
                   super_obj.is_array != new_obj.is_array or \
                   super_obj.array_size != new_obj.array_size or \
                   super_obj.embedded_object != new_obj.embedded_object:
                    mname = None
                    raise CIMError(
                        CIM_ERR_INVALID_PARAMETER,
                        _format("Invalid new_class parameter "
                                "{0!A} param {1|A}. "
                                "Does not match signature of "
                                "overridden method parameters "
                                "in class {2!A}.", mname, oname,
                                superclass.classname))
            else:
                assert True, "Invalid Type {0}" .format(type(super_obj))

            self._set_new_object(new_obj, super_obj,
                                 new_class,
                                 superclass, qualifier_repo,
                                 True, type_str)

            # if type is method, resolve the parameters.
            if isinstance(new_obj, CIMMethod):
                self._resolve_objects(
                    new_obj.parameters,
                    superclass_objects[new_obj.name].parameters,
                    new_class,
                    superclass, classrepo, qualifier_repo,
                    "Parameter", verbose=verbose)

        # Copy objects from from superclass that are not in new_class
        # Placed after loop with items in new_object so they are not part
        # of that loop.
        for oname, ovalue in six.iteritems(superclass_objects):
            if oname not in new_objects:
                new_value = ovalue.copy()
                new_value.propagated = True
                assert ovalue.class_origin
                new_value.class_origin = ovalue.class_origin
                for qualifier in new_value.qualifiers.values():
                    qualifier.propagated = True
                new_objects[oname] = new_value

    def _set_new_object(self, new_obj, inherited_obj, new_class, superclass,
                        qualifier_repo, propagated, type_str):
        """
        Set the object attributes for a single object and resolve the
        qualifiers. This sets attributes for Properties, Methods, and
        Parameters.
        """
        assert isinstance(new_obj, (CIMMethod, CIMProperty, CIMParameter))
        if inherited_obj:
            inherited_obj_qual = inherited_obj.qualifiers
        else:
            inherited_obj_qual = None

        if propagated:
            assert superclass is not None

        new_obj.propagated = propagated
        if propagated:
            assert inherited_obj is not None
            new_obj.class_origin = inherited_obj.class_origin
        else:
            assert inherited_obj is None
            new_obj.class_origin = new_class.classname
        self._resolve_qualifiers(new_obj.qualifiers,
                                 inherited_obj_qual,
                                 new_class,
                                 superclass,
                                 new_obj.name, type_str,
                                 qualifier_repo,
                                 propagate=propagated)

    def _resolve_qualifiers(self, new_quals, inherited_quals, new_class,
                            super_class, obj_name, obj_type, qualifier_repo,
                            propagate=False, verbose=False):
        """
        Process the override of qualifiers from the inherited_quals dictionary
        to the new_quals dict following the override rules in DSP0004.
        """
        superclassname = super_class.classname if super_class else None

        # TODO Diagnostic we will keep until really sure of this code
        if verbose:
            print("\nRESOLVE sc_name=%s nc_name=%s, obj_name=%s obj_type=%s "
                  "  propagate=%s" %
                  (superclassname, new_class.classname, obj_name, obj_type,
                   propagate))
            print('\nNEW QUAL')
            for q, qv in new_quals.items():
                print('  %s: %r' % (q, qv))
            print('INHERITEDQ:')
            if inherited_quals:
                for q, qv in inherited_quals.items():
                    print('  %s: %r' % (q, qv))

        # If propagate flag not set, initialize the qualfiers
        # by setting flavor defaults and propagated False
        if not propagate:
            for qname, qvalue in new_quals.items():
                self._init_qualifier(qvalue, qualifier_repo)
            return

        # resolve qualifiers not in inherited object
        for qname, qvalue in new_quals.items():
            if not inherited_quals or qname not in inherited_quals:
                self._init_qualifier(qvalue, qualifier_repo)

        # resolve qualifiers from inherited object
        for inh_qname, inh_qvalue in inherited_quals.items():
            if inh_qvalue.tosubclass:
                if inh_qvalue.overridable:
                    # if not in new quals, copy it to the new quals, else ignore
                    if inh_qname not in new_quals:
                        new_quals[inh_qname] = inherited_quals[inh_qname].copy()
                        new_quals[inh_qname].propagated = True
                    else:
                        new_quals[inh_qname].propagated = False
                        self._init_qualifier(new_quals[inh_qname],
                                             qualifier_repo)

                else:  # not overridable
                    if inh_qname in new_quals:
                        # allow for same qualifier def in subclass
                        # TODO should more than value match here.??
                        if new_quals[inh_qname].value != \
                                inherited_quals[inh_qname].value:
                            raise CIMError(
                                CIM_ERR_INVALID_PARAMETER,
                                _format("Invalid new_class {0!A}:{1!A} "
                                        "qualifier {2!A}. "
                                        "in class {3!A}. Not overridable ",
                                        obj_type, obj_name, inh_qname,
                                        new_class.classname))
                        else:
                            new_quals[inh_qname].propagated = True

                    else:  # not in new class, add it
                        new_quals[inh_qname] = inherited_quals[inh_qname].copy()
                        new_quals[inh_qname].propagated = True

            else:  # not tosubclass, i.e. restricted.
                if inh_qname in new_quals:
                    if inh_qvalue.overridable or inh_qvalue.overridable is None:
                        new_quals[inh_qname].propagated = True

                    else:
                        raise CIMError(
                            CIM_ERR_INVALID_PARAMETER,
                            _format("Invalid qualifier object {0!A} qualifier "
                                    "{1!A} . Restricted in super class {2!A}",
                                    obj_name, inh_qname, superclassname))

    def _resolve_class(self, new_class, namespace, qualifier_repo,
                       verbose=None):
        """
        Resolve the class defined by new_class by:

        1. Validating that the new class provided is a valid class.

        2. Validating the class against the repository to confirm that
           components required in the repository are in the repository, This
           includes the superclass if specified and the dependencies for
           EmbeddedInstance and reference properties.

        2. propagating any properties, methods, parameters, and qualifiers
           from the superclass if it is defined.

        """
        is_association_class = 'Association' in new_class.qualifiers

        if new_class.superclass:
            try:
                superclass = self._get_class(new_class.superclass,
                                             namespace=namespace,
                                             local_only=False,
                                             include_qualifiers=True,
                                             include_classorigin=True)
            except CIMError as ce:
                if ce.status_code == CIM_ERR_NOT_FOUND:
                    raise CIMError(
                        CIM_ERR_INVALID_SUPERCLASS,
                        _format("Superclass {0!A} for class {1!A} not found "
                                "in namespace {2!A}.",
                                new_class.superclass, new_class.classname,
                                namespace))
                else:
                    raise
        else:
            superclass = None

        # Validate association qualifier matches superclass
        if is_association_class and superclass:
            if 'Association' not in superclass.qualifiers:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("New class {0|A} derived from superclass {1!A} "
                            "in namespace {3!A} which is not Association "
                            "Class .",
                            new_class.classname, new_class.superclass,
                            namespace))

        # validate no reference properties in non-association
        for new_prop in six.itervalues(new_class.properties):
            if not is_association_class and new_prop.type == 'reference':
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format("Reference property {0!A} not allowed on "
                            "non-association class {1!A}",
                            new_prop.name, new_class.classname))

        objects = list(new_class.properties.values())
        for meth in new_class.methods.values():
            objects += list(meth.parameters.values())

        # Validate the attributes of all qualifiers in the new class
        if qualifier_repo:
            self._validate_qualifiers(new_class.qualifiers, qualifier_repo,
                                      new_class, 'CLASS')
            for pvalue in new_class.properties.values():
                self._validate_qualifiers(pvalue.qualifiers, qualifier_repo,
                                          new_class, 'PROPERTY')
            for mvalue in new_class.methods.values():
                self._validate_qualifiers(mvalue.qualifiers, qualifier_repo,
                                          new_class, 'METHOD')
                for pvalue in mvalue.parameters.values():
                    self._validate_qualifiers(pvalue.qualifiers, qualifier_repo,
                                              new_class, 'PARAMETER')

        # resolve class level qualifiers and attributes
        qualdict = superclass.qualifiers if superclass else {}
        new_class.classorigin = superclass.classname if superclass \
            else new_class.classname
        new_class.propagated = True if superclass else False
        self._resolve_qualifiers(new_class.qualifiers,
                                 qualdict,
                                 new_class,
                                 superclass,
                                 new_class.classname, 'class',
                                 qualifier_repo,
                                 propagate=False,
                                 verbose=verbose)

        classrepo = self._get_class_repo(namespace)
        # resolve properties in new class
        self._resolve_objects(new_class.properties,
                              superclass.properties if superclass else None,
                              new_class, superclass,
                              classrepo, qualifier_repo, "Property",
                              verbose=verbose)

        # resolve methods and parameters in new class
        self._resolve_objects(new_class.methods,
                              superclass.methods if superclass else None,
                              new_class, superclass,
                              classrepo, qualifier_repo, "Method",
                              verbose=verbose)

        return new_class



#!/usr/bin/env python

"""
Test exceptions module.
"""

from __future__ import absolute_import, print_function

import six
import pytest

from pywbem import Error, ConnectionError, AuthError, HTTPError, TimeoutError,\
    ParseError, CIMXMLParseError, XMLParseError, VersionError, CIMError, \
    ModelError, CIMInstance

# Test connection ID used for showing connection information in exception
# messages
TEST_CONN_ID = 'fake-conn-id'

# The expected connection information in exception messages
TEST_CONN_STR = 'Connection id: %s' % TEST_CONN_ID
TEST_CONN_STR_NONE = 'Connection id: None'


def _assert_subscription(exc):
    """
    Test the exception defined by exc for required args.
    """

    # Access by subscription is only supported in Python 2:
    if six.PY2:
        assert exc[:] == exc.args
        for i, _ in enumerate(exc.args):
            assert exc[i] == exc.args[i]
            assert exc[i:] == exc.args[i:]
            assert exc[0:i] == exc.args[0:i]
    else:
        try:
            _ = exc[:]
        except TypeError:
            pass
        else:
            assert False, "Access by slice did not fail in Python 3"
        for i, _ in enumerate(exc.args):
            try:
                _ = exc[i]  # noqa: F841
            except TypeError:
                pass
            else:
                assert False, "Access by index did not fail in Python 3"


def _assert_connection(exc, conn_id_kwarg, exp_conn_str):
    """
    Test the exception defined by exc for connection related information.
    """
    exp_conn_id = conn_id_kwarg.get('conn_id', None)
    assert exc.conn_id is exp_conn_id
    assert exc.conn_str == exp_conn_str


@pytest.fixture(params=[
    # The exception classes for which the simple test should be done:
    Error,
    ConnectionError,
    AuthError,
    TimeoutError,
    VersionError,
    ModelError,
], scope='module')
def simple_class(request):
    """
    Fixture representing variations of the simple exception classes.

    Returns the exception class.
    """
    return request.param


@pytest.fixture(params=[
    # Tuple of positional init arguments for the simple exception classes
    (),
    ('',),
    ('foo',),
    ('foo', 42),
], scope='module')
def simple_args(request):
    """
    Fixture representing variations of positional init arguments for the simple
    exception classes.

    Returns a tuple of positional arguments for initializing an exception
    object.
    """
    return request.param


@pytest.fixture(params=[
    # Tuple of (conn_id_kwarg, exp_conn_str)
    (dict(), TEST_CONN_STR_NONE),
    (dict(conn_id=None), TEST_CONN_STR_NONE),
    (dict(conn_id=TEST_CONN_ID), TEST_CONN_STR),
], scope='module')
def conn_info(request):
    """
    Fixture representing variations for the conn_id keyword argument for all
    exception classes, and the corresponding expected connection info string.

    Returns a tuple of:
    * conn_id_kwarg: dict with the 'conn_id' keyword argument. May be empty.
    * exp_conn_str: Expected connection info string.
    """
    return request.param


def test_simple(simple_class, simple_args, conn_info):
    # pylint: disable=redefined-outer-name
    """
    Test the simple exception classes.
    """

    conn_id_kwarg, exp_conn_str = conn_info

    exc = simple_class(*simple_args, **conn_id_kwarg)

    # exc has no len()
    assert len(exc.args) == len(simple_args)
    for i, _ in enumerate(simple_args):
        assert exc.args[i] == simple_args[i]
        assert exc.args[i:] == simple_args[i:]
        assert exc.args[0:i] == simple_args[0:i]
        assert exc.args[:] == simple_args[:]

    _assert_connection(exc, conn_id_kwarg, exp_conn_str)
    _assert_subscription(exc)


@pytest.fixture(params=[
    # The init arguments for the HTTPError exception class:
    # (status, reason, cimerror=None, cimdetails={})
    # Note: cimerror is the CIMError HTTP header field
    (200, 'OK'),
    (404, 'Not Found', 'instance xyz not found'),
    (404, 'Not Found', 'instance xyz not found', {'foo': 'bar'}),
], scope='module')
def httperror_args(request):
    """
    Fixture representing variations of positional init arguments for the
    HTTPError exception class.

    Returns a tuple of positional arguments for initializing a HTTPError
    exception object.
    """
    return request.param


def test_httperror(httperror_args, conn_info):
    # pylint: disable=redefined-outer-name
    """
    Test HTTPError exception class.
    """

    conn_id_kwarg, exp_conn_str = conn_info

    exc = HTTPError(*httperror_args, **conn_id_kwarg)

    assert exc.status == httperror_args[0]
    assert exc.reason == httperror_args[1]
    if len(httperror_args) < 3:
        assert exc.cimerror is None  # default value
    else:
        assert exc.cimerror == httperror_args[2]
    if len(httperror_args) < 4:
        assert exc.cimdetails == {}  # default value (set in init)
    else:
        assert exc.cimdetails == httperror_args[3]

    assert exc.args[0] == exc.status
    assert exc.args[1] == exc.reason
    assert exc.args[2] == exc.cimerror
    assert exc.args[3] == exc.cimdetails
    assert len(exc.args) == 4

    _assert_connection(exc, conn_id_kwarg, exp_conn_str)
    _assert_subscription(exc)


@pytest.fixture(params=[
    # The CIM status codes for the CIMError exception class:
    # (code, name)
    # Note: We don't test the exact default description text.
    # Note: name = None means that the status code is invalid.
    (0, None),
    (1, 'CIM_ERR_FAILED'),
    (2, 'CIM_ERR_ACCESS_DENIED'),
    (3, 'CIM_ERR_INVALID_NAMESPACE'),
    (4, 'CIM_ERR_INVALID_PARAMETER'),
    (5, 'CIM_ERR_INVALID_CLASS'),
    (6, 'CIM_ERR_NOT_FOUND'),
    (7, 'CIM_ERR_NOT_SUPPORTED'),
    (8, 'CIM_ERR_CLASS_HAS_CHILDREN'),
    (9, 'CIM_ERR_CLASS_HAS_INSTANCES'),
    (10, 'CIM_ERR_INVALID_SUPERCLASS'),
    (11, 'CIM_ERR_ALREADY_EXISTS'),
    (12, 'CIM_ERR_NO_SUCH_PROPERTY'),
    (13, 'CIM_ERR_TYPE_MISMATCH'),
    (14, 'CIM_ERR_QUERY_LANGUAGE_NOT_SUPPORTED'),
    (15, 'CIM_ERR_INVALID_QUERY'),
    (16, 'CIM_ERR_METHOD_NOT_AVAILABLE'),
    (17, 'CIM_ERR_METHOD_NOT_FOUND'),
    (18, None),
    (19, None),
    (20, 'CIM_ERR_NAMESPACE_NOT_EMPTY'),
    (21, 'CIM_ERR_INVALID_ENUMERATION_CONTEXT'),
    (22, 'CIM_ERR_INVALID_OPERATION_TIMEOUT'),
    (23, 'CIM_ERR_PULL_HAS_BEEN_ABANDONED'),
    (24, 'CIM_ERR_PULL_CANNOT_BE_ABANDONED'),
    (25, 'CIM_ERR_FILTERED_ENUMERATION_NOT_SUPPORTED'),
    (26, 'CIM_ERR_CONTINUATION_ON_ERROR_NOT_SUPPORTED'),
    (27, 'CIM_ERR_SERVER_LIMITS_EXCEEDED'),
    (28, 'CIM_ERR_SERVER_IS_SHUTTING_DOWN'),
    (29, None),
    (30, None),
], scope='module')
def status_tuple(request):
    """
    Fixture representing variations of CIM status codes for initializing a
    CIMError exception class.

    Returns a tuple of positional arguments for initializing a CIMError
    exception object.
    """
    return request.param


@pytest.fixture(params=[
    None,
    [],
    [CIMInstance('CIM_Err')],
    [CIMInstance('CIM_Err1'), CIMInstance('CIM_Err2')],
], scope='module')
def error_instances(request):
    """
    Fixture representing variations of the list of instances that can be
    set on a CIMError exception.

    Returns a list of CIMInstance objects, or None.
    """
    return request.param


def test_cimerror_1(status_tuple, conn_info):
    # pylint: disable=redefined-outer-name
    """
    Test CIMError exception class with just status_code as input.
    """

    status_code, status_code_name = status_tuple
    conn_id_kwarg, exp_conn_str = conn_info

    invalid_code_name = 'Invalid status code %s' % status_code
    invalid_code_desc = 'Invalid status code %s' % status_code

    exc = CIMError(status_code, **conn_id_kwarg)

    assert exc.status_code == status_code
    if status_code_name is None:
        assert exc.status_description == invalid_code_desc
        assert exc.status_code_name == invalid_code_name
    else:
        assert isinstance(exc.status_description, six.string_types)
        assert exc.status_code_name == status_code_name

    assert exc.args[0] == exc.status_code
    assert exc.args[1] is None
    assert exc.args[2] is None
    assert len(exc.args) == 3

    _assert_connection(exc, conn_id_kwarg, exp_conn_str)
    _assert_subscription(exc)


def test_cimerror_2(status_tuple, conn_info):
    # pylint: disable=redefined-outer-name
    """
    Test CIMError exception class with status_code and description as input.
    """

    status_code, status_code_name = status_tuple
    conn_id_kwarg, exp_conn_str = conn_info

    invalid_code_name = 'Invalid status code %s' % status_code
    input_desc = 'foo'

    exc = CIMError(status_code, input_desc, **conn_id_kwarg)

    assert exc.status_code == status_code
    assert exc.status_description == input_desc
    if status_code_name is None:
        assert exc.status_code_name == invalid_code_name
    else:
        assert exc.status_code_name == status_code_name

    assert exc.args[0] == exc.status_code
    assert exc.args[1] == input_desc
    assert exc.args[2] is None
    assert len(exc.args) == 3

    _assert_connection(exc, conn_id_kwarg, exp_conn_str)
    _assert_subscription(exc)


def test_cimerror_3(status_tuple, error_instances, conn_info):
    # pylint: disable=redefined-outer-name
    """
    Test CIMError exception class with status_code and instances as input.
    """

    status_code, status_code_name = status_tuple
    conn_id_kwarg, exp_conn_str = conn_info

    exc = CIMError(status_code, instances=error_instances, **conn_id_kwarg)

    assert exc.status_code == status_code
    if status_code_name is not None:
        assert exc.status_code_name == status_code_name

    assert exc.args[2] == error_instances
    assert len(exc.args) == 3

    _assert_connection(exc, conn_id_kwarg, exp_conn_str)
    _assert_subscription(exc)


@pytest.fixture(params=[
    # The exception classes for which the ParseError test should be done:
    ParseError,
    CIMXMLParseError,
    XMLParseError,
], scope='module')
def parseerror_class(request):
    """
    Fixture representing variations of the ParseError exception classes.

    Returns the exception class.
    """
    return request.param


@pytest.fixture(params=[
    # The init arguments for the ParseError exception class:
    # (message,)
    (None,),
    ('',),
    ('ill-formed XML',),
], scope='module')
def parseerror_args(request):
    """
    Fixture representing variations of positional init arguments for the
    ParseError exception class.

    Returns a tuple of positional arguments for initializing a ParseError
    exception object.
    """
    return request.param


def test_parseerror(parseerror_class, parseerror_args, conn_info):
    # pylint: disable=redefined-outer-name
    """
    Test ParseError exception class (and subclasses).
    """

    conn_id_kwarg, exp_conn_str = conn_info
    message = parseerror_args[0]

    exc = parseerror_class(*parseerror_args, **conn_id_kwarg)

    assert exc.args[0] == message
    assert len(exc.args) == 1

    _assert_connection(exc, conn_id_kwarg, exp_conn_str)
    _assert_subscription(exc)



#!/usr/bin/env python

"""Test WBEMListener against an indication generator. This test generates
   a number of indications and confirms that they are all received by the
   listener.

   This test validates the main paths of the listener and that the listener can
   receive large numbers of insdications without duplicates or dropping
   indications.

   It does not validate all of the possible xml options on indications.
"""

from __future__ import absolute_import

import unittest
import sys as _sys
import errno
import logging as _logging
from time import time
import datetime
from random import randint
import requests

from pywbem import WBEMListener

RCV_COUNT = 0
RCV_FAIL = False
VERBOSE = False
LISTENER = None


class ElapsedTimer(object):
    """
        Set up elapsed time timer. Calculates time between initiation
        and access.
    """

    def __init__(self):
        """ Initiate the object with current time"""
        self.start_time = datetime.datetime.now()

    def reset(self):
        """ Reset the start time for the timer"""
        self.start_time = datetime.datetime.now()

    def elapsed_ms(self):
        """ Get the elapsed time in milliseconds. returns floating
            point representation of elapsed time in seconds.
        """
        dt = datetime.datetime.now() - self.start_time
        return ((dt.days * 24 * 3600) + dt.seconds) * 1000 + \
            dt.microseconds / 1000.0

    def elapsed_sec(self):
        """ get the elapsed time in seconds. Returns floating
            point representation of time in seconds
        """
        return self.elapsed_ms() / 1000


def create_indication_data(msg_id, sequence_number, delta_time, protocol_ver):
    """Create a test indication from the template and input attributes"""

    data_template = """<?xml version="1.0" encoding="utf-8" ?>
    <CIM CIMVERSION="2.0" DTDVERSION="2.4">
      <MESSAGE ID="%(msg_id)s" PROTOCOLVERSION="%(protocol_ver)s">
        <SIMPLEEXPREQ>
          <EXPMETHODCALL NAME="ExportIndication">
            <EXPPARAMVALUE NAME="NewIndication">
              <INSTANCE CLASSNAME="CIM_AlertIndication">
                <PROPERTY NAME="Severity" TYPE="string">
                  <VALUE>high</VALUE>
                </PROPERTY>
                <PROPERTY NAME="SequenceNumber" TYPE="string">
                  <VALUE>%(sequence_number)s</VALUE>
                </PROPERTY>
                <PROPERTY NAME="DELTA_TIME" TYPE="string">
                  <VALUE>%(delta_time)s</VALUE>
                </PROPERTY>
              </INSTANCE>
            </EXPPARAMVALUE>
          </EXPMETHODCALL>
        </SIMPLEEXPREQ>
      </MESSAGE>
    </CIM>"""

    data = {'sequence_number': sequence_number, 'delta_time': delta_time,
            'protocol_ver': protocol_ver, 'msg_id': msg_id}
    return data_template % data


def send_indication(url, headers, payload, verbose):
    """Send a single indication using Python requests"""

    try:
        response = requests.post(url, headers=headers, data=payload, timeout=4)
    except Exception as ex:  # pylint: disable=broad-except
        print('Exception %s' % ex)
        return False

    if verbose:
        print('\nResponse code=%s headers=%s data=%s' % (response.status_code,
                                                         response.headers,
                                                         response.text))

    return True if(response.status_code == 200) else False


def _process_indication(indication, host):
    """
    This function gets called when an indication is received.
    It receives each indication on a separate thread so the only communication
    with the rest of the program is RCV_COUNT which it increments for each
    received indication

    It tests the received indication sequence number against the RCV_COUNT
    which should catch and indication duplication or missing since the counters
    would no longer match.

    NOTE: Since this is a standalone function, it does not do an assert fail
    if there is a mismatch.
    """

    global RCV_COUNT  # pylint: disable=global-statement
    global RCV_FAIL  # pylint: disable=global-statement

    counter = indication.properties['SequenceNumber'].value
    if int(counter) != RCV_COUNT:
        RCV_FAIL = True
        print('ERROR in process indication counter=%s, COUNT=%s' % (counter,
                                                                    RCV_COUNT))
    # print('counter %s COUNT %s' % (counter, RCV_COUNT))
    RCV_COUNT += 1
    if VERBOSE:
        print("Received indication %s from %s:\n%s" % (RCV_COUNT, host,
                                                       indication.tomof()))


class TestIndications(unittest.TestCase):
    """
    Create a WBEMListener and starts the listener. Note that it resets the
    received indication counter (RCV_COUNT) so that there is an accurate
    count of indications actually received.
    """

    @staticmethod
    def createlistener(host, http_port=None, https_port=None,
                       certfile=None, keyfile=None):
        """
        Create and start a listener based on host, ports, etc.
        """
        global RCV_COUNT  # pylint: disable=global-statement
        global LISTENER  # pylint: disable=global-statement
        global RCV_FAIL  # pylint: disable=global-statement
        RCV_FAIL = False  # pylint: disable=global-statement

        _logging.basicConfig(stream=_sys.stderr, level=_logging.WARNING,
                             format='%(levelname)s: %(message)s')

        RCV_COUNT = 0
        LISTENER = WBEMListener(host=host,
                                http_port=http_port,
                                https_port=https_port,
                                certfile=certfile,
                                keyfile=keyfile)
        LISTENER.add_callback(_process_indication)
        LISTENER.start()

    # pylint: disable=unused-argument
    def send_indications(self, send_count, http_port):
        """
        Send the number of indications defined by the send_count attribute
        using the specified listener HTTP port.

        Creates the listener, starts the listener, creates the
        indication XML and adds sequence number and time to the
        indication instance and sends that instance using requests.
        The indication instance is modified for each indication count so
        that each carries its own sequence number.
        """

        # pylint: disable=global-variable-not-assigned
        global VERBOSE  # pylint: disable=global-statement
        global RCV_FAIL  # pylint: disable=global-statement
        RCV_FAIL = False
        host = 'localhost'
        try:
            self.createlistener(host, http_port)

            start_time = time()

            full_url = 'http://%s:%s' % (host, http_port)

            cim_protocol_version = '1.4'

            headers = {'content-type': 'application/xml; charset=utf-8',
                       'CIMExport': 'MethodRequest',
                       'CIMExportMethod': 'ExportIndication',
                       'Accept-Encoding': 'Identity',
                       'CIMProtocolVersion': cim_protocol_version}
            # We include accept-encoding because of requests issue.
            # He supplies it if we don't.  TODO try None

            delta_time = time() - start_time
            rand_base = randint(1, 10000)
            timer = ElapsedTimer()
            for i in range(send_count):

                msg_id = '%s' % (i + rand_base)
                payload = create_indication_data(msg_id, i, delta_time,
                                                 cim_protocol_version)

                if VERBOSE:
                    print('headers=%s\n\npayload=%s' % (headers, payload))

                success = send_indication(full_url, headers, payload, VERBOSE)

                if success:
                    if VERBOSE:
                        print('sent # %s' % i)
                else:
                    self.fail('Error return from send. Terminating.')

            endtime = timer.elapsed_sec()
            if VERBOSE:
                print('Sent %s indications in %s sec or %.2f ind/sec' %
                      (send_count, endtime, (send_count / endtime)))

            self.assertEqual(send_count, RCV_COUNT,
                             'Mismatch between sent and rcvd')
            self.assertFalse(RCV_FAIL, 'Sequence numbers do not match')
            RCV_FAIL = False
        finally:
            LISTENER.stop()

    def test_send_10(self):
        """Test with sending 10 indications"""
        self.send_indications(10, 50000)

    def test_send_100(self):
        """Test sending 100 indications"""
        self.send_indications(100, 50000)

    # Disabled the following tests, because in some environments it takes 30min.
    # def test_send_1000(self):
    #     """Test sending 1000 indications"""
    #     self.send_indications(1000, 50000)

    def test_attrs(self):
        # pylint: disable=no-self-use
        """
        Test WBEMListener attributes.
        """

        host = 'localhost'
        http_port = '50000'  # as a string
        exp_http_port = 50000  # as an integer

        listener = WBEMListener(host, http_port)
        assert listener.host == host
        assert listener.http_port == exp_http_port
        assert listener.https_port is None
        assert listener.certfile is None
        assert listener.keyfile is None
        assert isinstance(listener.logger, _logging.Logger)
        assert listener.http_started is False
        assert listener.https_started is False

    def test_start_stop(self):
        # pylint: disable=no-self-use
        """
        Test starting and stopping of the the listener.
        """

        host = 'localhost'
        http_port = '50000'

        listener = WBEMListener(host, http_port)
        assert listener.http_started is False
        assert listener.https_started is False

        listener.start()
        assert listener.http_started is True
        assert listener.https_started is False

        listener.stop()
        assert listener.http_started is False
        assert listener.https_started is False

    def test_port_in_use(self):
        # pylint: disable=no-self-use
        """
        Test starting the listener when port is in use by another listener.
        """

        host = 'localhost'

        # Don't use this port in other tests, to be on the safe side
        # as far as port reuse is concerned.
        http_port = '59999'

        exp_exc_type = OSError

        listener1 = WBEMListener(host, http_port)
        listener1.start()
        assert listener1.http_started is True

        listener2 = WBEMListener(host, http_port)

        try:

            # The code to be tested
            listener2.start()

        except Exception as exc:  # pylint: disable=broad-except
            # e.g. on Linux
            assert isinstance(exc, exp_exc_type)
            assert getattr(exc, 'errno', None) == errno.EADDRINUSE
            assert listener2.http_started is False
        else:
            # e.g. on Windows
            assert listener2.http_started is True

        # Verify that in any case, listener1 is still started
        assert listener1.http_started is True

        listener1.stop()  # cleanup
        listener2.stop()  # cleanup (for cases where it started)

    def test_context_mgr(self):
        # pylint: disable=no-self-use
        """
        Test starting the listener and automatic closing in a context manager.
        """

        host = 'localhost'

        # Don't use this port in other tests, to be on the safe side
        # as far as port reuse is concerned.
        http_port = '59998'

        # The code to be tested (is the context manager)
        with WBEMListener(host, http_port) as listener1:

            # Verify that CM enter returns the listener
            assert isinstance(listener1, WBEMListener)

            listener1.start()
            assert listener1.http_started is True

        # Verify that CM exit stops the listener
        assert listener1.http_started is False

        # Verify that the TCP/IP port can be used again
        listener2 = WBEMListener(host, http_port)
        listener2.start()
        assert listener2.http_started is True
        listener2.stop()


if __name__ == '__main__':
    VERBOSE = False
    unittest.main()



#!/usr/bin/env python

import sys
try:
    from pbr.version import VersionInfo
except ImportError:
    # Version cannot be determined
    sys.exit(0)

if len(sys.argv) != 2:
    print("Error: Incorrect number of command line arguments")
    print("Usage: {0}  package_name".format(sys.argv[0]))
    sys.exit(2)

pkg_name = sys.argv[1]
pkg_version = VersionInfo(pkg_name).release_string()
print(pkg_version)



#!/usr/bin/env python
"""
Test cases for tupletree module and Unicode/XML check functions.
"""

from __future__ import absolute_import, print_function

import xml
import re
import six
import pytest

from pywbem import tupletree, ParseError


# The following was originally the pywbem parser.  Now it is used simply as a
# check on the SAX parser.  Therefore, this code was moved from tupletree.py to
# here for the tests. Note that this code is referenced from other code in the
# attic so we might not want to eliminate it completely.

def dom_to_tupletree(node):
    """
    Convert a DOM object to a pyRXP-style tuple tree.

    Each element is a 4-tuple of (NAME, ATTRS, CONTENTS, None).

    Very nice for processing complex nested trees.
    """

    if node.nodeType == node.DOCUMENT_NODE:
        # boring; pop down one level
        return dom_to_tupletree(node.firstChild)
    assert node.nodeType == node.ELEMENT_NODE

    name = node.nodeName
    attrs = {}
    contents = []

    for child in node.childNodes:
        if child.nodeType == child.ELEMENT_NODE:
            contents.append(dom_to_tupletree(child))
        elif child.nodeType == child.TEXT_NODE:
            assert isinstance(child.nodeValue, six.string_types), \
                "text node %s is not a string %r" % child
            contents.append(child.nodeValue)
        elif child.nodeType == child.CDATA_SECTION_NODE:
            contents.append(child.nodeValue)
        else:
            raise RuntimeError("can't handle %r" % child)

    for i in range(node.attributes.length):
        attr_node = node.attributes.item(i)
        attrs[attr_node.nodeName] = attr_node.nodeValue

    # TODO: Cannot handle comments, cdata, processing instructions, etc.

    # it's so easy in retrospect!
    return (name, attrs, contents)


def xml_to_tupletree(xml_string):
    """
    Parse XML straight into tupletree.
    Uses the minidom to parse xml_string int a dom object.
    This is part of the old obsolete dom code
    """
    dom_xml = xml.dom.minidom.parseString(xml_string)
    return dom_to_tupletree(dom_xml)


class Test_xml_to_tupletree_sax(object):
    # pylint: disable=too-few-public-methods
    """
    Exhaustive tests for tupletree.xml_to_tupletree_sax(), with inline input
    XML and expected tupletrees (for success) or exceptions (for failures).

    A comparison with the result of the previously implemented DOM based
    pywbem parser ensures that the SAX parsing is compatible.
    """

    testcases = [
        # Each testcase has these items:
        # * desc: Short testcase description.
        # * xml_string: XML string to be parsed, as unicode string.
        # * exp_tupletree: Expected tupletree parse result,
        #   or None if expecting failure.
        # * exp_exc_type: Expected exception type,
        #   or None if expecting success.
        # * exp_exc_msg_pattern: Expected pattern in exception message,
        #   or None if expecting success.
        # * condition: Condition for testcase to run.

        # General good cases
        (
            "One-line XML string with one empty element",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<TEXT></TEXT>',
            (u'TEXT',
             {},
             []),
            None, None,
            True
        ),
        (
            "One-line XML string with one empty element, short form",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<TEXT/>',
            (u'TEXT',
             {},
             []),
            None, None,
            True
        ),
        (
            "One-line XML string with one element with ASCII text",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<TEXT>foo</TEXT>',
            (u'TEXT',
             {},
             [u'foo']),
            None, None,
            True
        ),
        (
            "One-line XML string with one element with non-ASCII UCS-2 text",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<TEXT>\u00e8</TEXT>',
            (u'TEXT',
             {},
             [u'\u00e8']),
            None, None,
            True
        ),
        (
            "One-line XML string with one element with non-UCS-2 text",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<TEXT>\U00010142</TEXT>',  # GREEK ACROPHONIC ATTIC ONE DRACHMA
            (u'TEXT',
             {},
             [u'\U00010142']),
            None, None,
            True
        ),
        (
            "One-line XML string with one element with one attribute",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<TEXT ATTR="foo">bar</TEXT>',
            (u'TEXT',
             {u'ATTR': u'foo'},
             [u'bar']),
            None, None,
            True
        ),
        (
            "One-line XML string with one element with two attributes",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<TEXT ATTR1="foo1" ATTR2="foo2">bar</TEXT>',
            (u'TEXT',
             {u'ATTR1': u'foo1', u'ATTR2': u'foo2'},
             [u'bar']),
            None, None,
            True
        ),
        (
            "One-line XML string with two elements and two attributes",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<TEXT1 ATTR1="foo1"><TEXT2 ATTR2="foo2">bar</TEXT2></TEXT1>',
            (u'TEXT1',
             {u'ATTR1': u'foo1'},
             [(u'TEXT2',
               {u'ATTR2': u'foo2'},
               [u'bar'])
              ]),
            None, None,
            True
        ),

        # Multi-line good cases
        (
            "Multi-line XML string, text without newline",
            u'<?xml version="1.0" encoding="utf-8" ?>\n'
            u'<TEXT1 ATTR1="foo1">\n'
            u'<TEXT2 ATTR2="foo2">bar</TEXT2>\n'
            u'</TEXT1>',
            (u'TEXT1',
             {u'ATTR1': u'foo1'},
             [u'\n',
              (u'TEXT2',
               {u'ATTR2': u'foo2'},
               [u'bar']),
              u'\n']),
            None, None,
            True
        ),
        (
            "Multi-line XML string, text with newlines",
            u'<?xml version="1.0" encoding="utf-8" ?>\n'
            u'<TEXT1 ATTR1="foo1">\n'
            u'<TEXT2 ATTR2="foo2">\n'
            u'bar\n'
            u'</TEXT2>\n'
            u'</TEXT1>',
            (u'TEXT1',
             {u'ATTR1': u'foo1'},
             [u'\n',
              (u'TEXT2',
               {u'ATTR2': u'foo2'},
               [u'\nbar\n']),
              u'\n']),
            None, None,
            True
        ),

        # General failure cases
        (
            "Empty XML string",
            u'',
            None,
            ParseError, "XML parsing error.*no element found",
            True
        ),
        (
            "Just the XML prolog",
            u'<?xml version="1.0" encoding="utf-8" ?>',
            None,
            ParseError, "XML parsing error.*no element found",
            True
        ),
        (
            "Start element without end element",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<ELEM>',
            None,
            ParseError, "XML parsing error.*no element found",
            True
        ),
        (
            "End element misses trailing angle bracket",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<ELEM>abc</ELEM',
            None,
            ParseError, "XML parsing error.*unclosed token",
            True
        ),
        (
            "End element misses leading angle bracket",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<ELEM>abc/ELEM>',
            None,
            ParseError, "XML parsing error.*no element found",
            True
        ),
        (
            "Start element misses trailing angle bracket",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<ELEM</ELEM>',
            None,
            ParseError, "XML parsing error.*not well-formed",
            True
        ),
        (
            "Start element misses leading angle bracket",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'ELEM></ELEM>',
            None,
            ParseError, "XML parsing error.*syntax error",
            True
        ),

        # Failure cases with line number checking
        (
            "Failing multi-line XML string with failure in line 3",
            u'<?xml version="1.0" encoding="utf-8" ?>\n'
            u'<ELEM1>\n'
            u'<ELEM2</ELEM2>\n'
            u'<ELEM1>\n',
            None,
            ParseError,
            "XML parsing error.*not well-formed.*"
            "Line 3 column 6 of XML string.*"
            "<ELEM2</ELEM2>",
            True
        ),

        # Failure cases with non-ASCII chars
        (
            "Failing XML string with non-ASCII attribute value",
            u'<?xml version="1.0" encoding="utf-8" ?>'
            u'<ELEM1 ATTR1="\u00e8"></ELEM2>',
            None,
            ParseError,
            "XML parsing error.*mismatched tag.*"
            "Line 1 column 58 of XML string.*"
            r'<ELEM1 ATTR1="\\xc3\\xa8"></ELEM2>',
            True
        ),

        # CIM-XML cases
        (
            "CIM-XML response with MESSAGE end element missing "
            "(former Associators_error.xml file)",
            u'<?xml version="1.0" encoding="utf-8" ?>\n'
            u'<CIM CIMVERSION="2.0" DTDVERSION="2.0">\n'
            u'  <MESSAGE ID="1001" PROTOCOLVERSION="1.0">\n'
            u'    <SIMPLERSP>\n'
            u'      <IMETHODRESPONSE NAME="Associators">\n'
            u'        <IRETURNVALUE>\n'
            u'        </IRETURNVALUE>\n'
            u'      </IMETHODRESPONSE>\n'
            u'    </SIMPLERSP>\n'
            u'</CIM>\n',
            None,
            ParseError, "XML parsing error.*mismatched tag",
            True
        ),
        (
            "CIM-XML response from Associators operation "
            "(former Associators_Empty.xml)",
            u'<?xml version="1.0" encoding="utf-8" ?>\n'
            u'<CIM CIMVERSION="2.0" DTDVERSION="2.0">\n'
            u'  <MESSAGE ID="1001" PROTOCOLVERSION="1.0">\n'
            u'    <SIMPLERSP>\n'
            u'      <IMETHODRESPONSE NAME="Associators">\n'
            u'        <IRETURNVALUE>\n'
            u'        </IRETURNVALUE>\n'
            u'      </IMETHODRESPONSE>\n'
            u'    </SIMPLERSP>\n'
            u'  </MESSAGE>\n'
            u'</CIM>\n',
            (u'CIM',
             {u'CIMVERSION': u'2.0', u'DTDVERSION': u'2.0'},
             [u'\n  ',
              (u'MESSAGE',
               {u'ID': u'1001', u'PROTOCOLVERSION': u'1.0'},
               [u'\n    ',
                (u'SIMPLERSP',
                 {},
                 [u'\n      ',
                  (u'IMETHODRESPONSE',
                   {u'NAME': u'Associators'},
                   [u'\n        ',
                    (u'IRETURNVALUE',
                     {},
                     [u'\n        ']),
                    u'\n      ']),
                  u'\n    ']),
                u'\n  ']),
              u'\n']),
            None, None,
            True
        ),
        (
            "CIM-XML response from Associators operation "
            "(former Associators_StorageVolume_small.xml)",
            u'<?xml version="1.0" encoding="utf-8" ?>\n'
            u'<CIM CIMVERSION="2.0" DTDVERSION="2.0"><MESSAGE ID="1001" '
            u'PROTOCOLVERSION="1.0">\n'
            u'<SIMPLERSP><IMETHODRESPONSE NAME="Associators"><IRETURNVALUE>\n'
            u'    \n'
            u'<VALUE.OBJECTWITHPATH>\n'
            u'<INSTANCEPATH>\n'
            u'<NAMESPACEPATH>\n'
            u'<HOST>10.10.10.10</HOST>\n'
            u'<LOCALNAMESPACEPATH>\n'
            u'<NAMESPACE NAME="root"/>\n'
            u'<NAMESPACE NAME="emc"/>\n'
            u'</LOCALNAMESPACEPATH>\n'
            u'</NAMESPACEPATH>\n'
            u'<INSTANCENAME CLASSNAME="Symm_StorageVolume">\n'
            u'<KEYBINDING NAME="CreationClassName"><KEYVALUE '
            u'VALUETYPE="string" TYPE="string">Symm_StorageVolume</KEYVALUE>'
            u'</KEYBINDING>\n'
            u'<KEYBINDING NAME="DeviceID"><KEYVALUE VALUETYPE="string" '
            u'TYPE="string">00000</KEYVALUE></KEYBINDING>\n'
            u'<KEYBINDING NAME="SystemCreationClassName"><KEYVALUE '
            u'VALUETYPE="string" TYPE="string">Symm_StorageSystem</KEYVALUE>'
            u'</KEYBINDING>\n'
            u'<KEYBINDING NAME="SystemName"><KEYVALUE VALUETYPE="string" '
            u'TYPE="string">SYMMETRIX-+-000194900000</KEYVALUE></KEYBINDING>\n'
            u'</INSTANCENAME>\n'
            u'</INSTANCEPATH>\n'
            u'<INSTANCE CLASSNAME="Symm_StorageVolume">\n'
            u'<PROPERTY NAME="Usage" TYPE="uint16"><VALUE>3</VALUE>\n'
            u'</PROPERTY>\n'
            u'</INSTANCE>\n'
            u'</VALUE.OBJECTWITHPATH>\n'
            u'\n'
            u'</IRETURNVALUE></IMETHODRESPONSE></SIMPLERSP></MESSAGE></CIM>\n',
            (u'CIM',
             {u'CIMVERSION': u'2.0', u'DTDVERSION': u'2.0'},
             [(u'MESSAGE',
               {u'ID': u'1001', u'PROTOCOLVERSION': u'1.0'},
               [u'\n',
                (u'SIMPLERSP',
                 {},
                 [(u'IMETHODRESPONSE',
                   {u'NAME': u'Associators'},
                   [(u'IRETURNVALUE',
                     {},
                     [u'\n    \n',
                      (u'VALUE.OBJECTWITHPATH',
                       {},
                       [u'\n',
                        (u'INSTANCEPATH',
                         {},
                         [u'\n',
                          (u'NAMESPACEPATH',
                           {},
                           [u'\n',
                            (u'HOST',
                             {},
                             [u'10.10.10.10']),
                            u'\n',
                            (u'LOCALNAMESPACEPATH',
                             {},
                             [u'\n',
                              (u'NAMESPACE',
                               {u'NAME': u'root'},
                               []),
                              u'\n',
                              (u'NAMESPACE',
                               {u'NAME': u'emc'},
                               []),
                              u'\n']),
                            u'\n']),
                          u'\n',
                          (u'INSTANCENAME',
                           {u'CLASSNAME': u'Symm_StorageVolume'},
                           [u'\n',
                            (u'KEYBINDING',
                             {u'NAME': u'CreationClassName'},
                             [(u'KEYVALUE',
                               {u'TYPE': u'string',
                                u'VALUETYPE': u'string'},
                               [u'Symm_StorageVolume'])
                              ]),
                            u'\n',
                            (u'KEYBINDING',
                             {u'NAME': u'DeviceID'},
                             [(u'KEYVALUE',
                               {u'TYPE': u'string',
                                u'VALUETYPE': u'string'},
                               [u'00000'])
                              ]),
                            u'\n',
                            (u'KEYBINDING',
                             {u'NAME': u'SystemCreationClassName'},
                             [(u'KEYVALUE',
                               {u'TYPE': u'string',
                                u'VALUETYPE': u'string'},
                               [u'Symm_StorageSystem'])
                              ]),
                            u'\n',
                            (u'KEYBINDING',
                             {u'NAME': u'SystemName'},
                             [(u'KEYVALUE',
                               {u'TYPE': u'string',
                                u'VALUETYPE': u'string'},
                               [u'SYMMETRIX-+-000194900000'])
                              ]),
                            u'\n']),
                          u'\n']),
                        u'\n',
                        (u'INSTANCE',
                         {u'CLASSNAME': u'Symm_StorageVolume'},
                         [u'\n',
                          (u'PROPERTY',
                           {u'NAME': u'Usage',
                            u'TYPE': u'uint16'},
                           [(u'VALUE',
                             {},
                             [u'3']),
                            u'\n']),
                          u'\n']),
                        u'\n']),
                      u'\n\n'])
                    ])
                  ])
                ])
              ]),
            None, None,
            True
        ),
    ]

    @pytest.mark.parametrize(
        "desc, xml_string, exp_tupletree, exp_exc_type, exp_exc_msg_pattern, "
        "condition",
        testcases
    )
    @pytest.mark.parametrize(
        "encoding",
        ['unicode', 'bytes']
    )
    def test_xml_to_tupletree_sax(
            self, encoding, desc, xml_string, exp_tupletree, exp_exc_type,
            exp_exc_msg_pattern, condition):
        # pylint: disable=no-self-use
        """
        Test xml_to_tupletree_sax() against expected results and against
        DOM parser results.
        """

        if not condition:
            pytest.skip("Condition for test case not met")

        assert isinstance(xml_string, six.text_type)
        if encoding == 'bytes':
            xml_string = xml_string.encode("utf-8")
        else:
            assert encoding == 'unicode'

        if exp_exc_type is None:

            act_tupletree = tupletree.xml_to_tupletree_sax(
                xml_string, 'Test XML')

            # Compare against expected result
            assert act_tupletree == exp_tupletree

            if encoding == 'bytes':
                # Compare against result from previously used DOM parser
                dom_tupletree = xml_to_tupletree(xml_string)
                assert act_tupletree == dom_tupletree

        else:
            with pytest.raises(exp_exc_type) as exec_info:

                tupletree.xml_to_tupletree_sax(xml_string, 'Test XML')

            if exp_exc_msg_pattern:
                exc = exec_info.value
                exc_msg = str(exc)
                one_line_exc_msg = exc_msg.replace('\n', '\\n')
                assert re.search(exp_exc_msg_pattern, one_line_exc_msg), \
                    "Unexpected exception message:\n" + exc_msg


class Test_get_failing_line(object):
    # pylint: disable=too-few-public-methods
    """Tests for tupletree.get_failing_line()"""

    testcases = [
        # Each testcase has these items:
        # * desc: Short testcase description.
        # * xml_string: Input XML string for the function.
        # * exc_msg: Input exception message for the function.
        # * exp_result: Expected result tuple, or None if expecting failure.
        # * exp_exc_type: Expected exception type, or None if expecting success.
        # * condition: Condition for testcase to run.

        # General cases
        (
            "Short one line XML string, parseable exception message",
            b'<V>ab</V>',
            ":1:2: some message",
            (1, 2, 3, "'<V>ab</V>'"),
            None, True
        ),
        (
            "Short 2 line XML string, parseable exception message",
            b'<V>ab</V>\n'
            b'<X>ab</X>\n',
            ":2:2: some message",
            (2, 2, 3, "'<X>ab</X>'"),
            None, True
        ),
        (
            "Short 2 line XML string, non-parseable exception message",
            b'<V>ab</V>\n'
            b'<X>ab</X>\n',
            "(2:2) some message",
            (None, None, None, "'<V>ab</V>\\n<X>ab</X>\\n'"),
            None, True
        ),
        (
            "Short 2 line XML string, parseable exception message for line 3",
            b'<V>ab</V>\n'
            b'<X>ab</X>\n',
            ":3:2: some message",
            (None, None, None, "'<V>ab</V>\\n<X>ab</X>\\n'"),
            None, True
        ),
        (
            "Empty XML, parseable exception message for line 1",
            b'',
            ":1:1: empty string",
            (1, 1, 2, "''"),
            None, True
        ),
        (
            "Empty line with newline XML, parseable exc. message for line 1",
            b'\n',
            ":1:1: empty string",
            (1, 1, 2, "''"),
            None, True
        ),

        # Exceeding chars after colno, with 0 chars before colno
        (
            "XML with 0/x/500 chars",
            b'b' * 0 + b'x' + b'a' * 500,
            ":1:1: empty string",
            (1, 1, 2, "'" + 'b' * 0 + 'x' + 'a' * 500 + "'"),
            None, True
        ),
        (
            "XML with 0/x/501 chars",
            b'b' * 0 + b'x' + b'a' * 501,
            ":1:1: empty string",
            (1, 1, 2, "'" + 'b' * 0 + 'x' + 'a' * 500 + "'" + '...'),
            None, True
        ),

        # Exceeding chars after colno, with 5 chars before colno
        (
            "XML with 5/x/500 chars",
            b'b' * 5 + b'x' + b'a' * 500,
            ":1:6: empty string",
            (1, 6, 7, "'" + 'b' * 5 + 'x' + 'a' * 500 + "'"),
            None, True
        ),
        (
            "XML with 5/x/501 chars",
            b'b' * 5 + b'x' + b'a' * 501,
            ":1:6: empty string",
            (1, 6, 7, "'" + 'b' * 5 + 'x' + 'a' * 500 + "'" + '...'),
            None, True
        ),

        # Exceeding chars before colno, with 0 chars after colno
        (
            "XML with 500/x/0 chars",
            b'b' * 500 + b'x' + b'a' * 0,
            ":1:501: empty string",
            (1, 501, 502, "'" + 'b' * 500 + 'x' + 'a' * 0 + "'"),
            None, True
        ),
        (
            "XML with 501/x/0 chars",
            b'b' * 501 + b'x' + b'a' * 0,
            ":1:502: empty string",
            (1, 502, 505, '...' + "'" + 'b' * 500 + 'x' + 'a' * 0 + "'"),
            None, True
        ),

        # Exceeding chars before colno, with 5 chars after colno
        (
            "XML with 500/x/5 chars",
            b'b' * 500 + b'x' + b'a' * 5,
            ":1:501: empty string",
            (1, 501, 502, "'" + 'b' * 500 + 'x' + 'a' * 5 + "'"),
            None, True
        ),
        (
            "XML with 501/x/5 chars",
            b'b' * 501 + b'x' + b'a' * 5,
            ":1:502: empty string",
            (1, 502, 505, '...' + "'" + 'b' * 500 + 'x' + 'a' * 5 + "'"),
            None, True
        ),

        # Exceeding chars before and after colno
        (
            "XML with 500/x/500 chars",
            b'b' * 500 + b'x' + b'a' * 500,
            ":1:501: empty string",
            (1, 501, 502, "'" + 'b' * 500 + 'x' + 'a' * 500 + "'"),
            None, True
        ),
        (
            "XML with 501/x/501 chars",
            b'b' * 501 + b'x' + b'a' * 501,
            ":1:502: empty string",
            (1, 502, 505, '...' + "'" + 'b' * 500 + 'x' + 'a' * 500 + "'" +
             '...'),
            None, True
        ),

    ]

    @pytest.mark.parametrize(
        "desc, xml_string, exc_msg, exp_result, exp_exc_type, condition",
        testcases
    )
    def test_get_failing_line(
            self, desc, xml_string, exc_msg, exp_result, exp_exc_type,
            condition):
        # pylint: disable=no-self-use
        """Tests for tupletree.get_failing_line()"""

        if not condition:
            pytest.skip("Condition for test case not met")

        if exp_exc_type is None:

            act_result = tupletree.get_failing_line(xml_string, exc_msg)

            assert isinstance(act_result, tuple)
            assert act_result == exp_result

        else:
            with pytest.raises(exp_exc_type):

                tupletree.get_failing_line(xml_string, exc_msg)


class Test_check_invalid_utf8_sequences(object):
    # pylint: disable=too-few-public-methods
    """
    Tests for tupletree.check_invalid_utf8_sequences()
    """

    testcases = [
        # Each testcase has these items:
        # * desc: Short testcase description.
        # * utf8_string: Input string for the function.
        # * exp_exc_type: Expected exception type, or None if expecting success.
        # * condition: Condition for testcase to run.

        # General cases
        (
            "Incorrect type of input string",
            u'<V>ab</V>',
            TypeError, True
        ),

        # Good cases
        (
            "Good case without special chars",
            b'<V>a</V>',
            None, True
        ),
        (
            "Good case with U+0009 (TAB), U+000A (NL), U+000D (CR) chars",
            b'<V>a\x09b\x0Ac\x0Dd</V>',
            None, True
        ),
        (
            "Good case with U+0350 char (in UTF-8)",
            b'<V>a\xCD\x90b</V>',
            None, True
        ),
        (
            "Good case with U+2013 char (in UTF-8)",
            b'<V>a\xE2\x80\x93b</V>',
            None, True
        ),
        (
            "Good case with U+10122 char (in UTF-8)",
            b'<V>a\xF0\x90\x84\xA2b</V>',
            None, True
        ),

        # Correctly encoded but ill-formed UTF-8
        (
            "Ill-formed UTF-8 using surrogate U+D800,U+DD22 "
            "(in otherwise correctly encoded UTF-8)",
            b'<V>a\xED\xA0\x80\xED\xB4\xA2b</V>',
            ParseError, True
        ),
        (
            "Ill-formed UTF-8 using surrogates U+D800,U+DD22 and U+D800,U+DD23 "
            "(in otherwise correctly encoded UTF-8)",
            b'<V>a\xED\xA0\x80\xED\xB4\xA2b\xED\xA0\x80\xED\xB4\xA3</V>',
            ParseError, True
        ),

        # Incorrectly encoded UTF-8
        (
            "Incorrectly encoded UTF-8 with 1-byte sequence",
            b'<V>a\x80b</V>',
            ParseError, True
        ),
        (
            "Incorrectly encoded UTF-8 with 2-byte sequence "
            "with missing second byte",
            b'<V>a\xC0',
            ParseError, True
        ),
        (
            "Incorrectly encoded UTF-8 with 2-byte sequence "
            "with incorrect 2nd byte",
            b'<V>a\xC0b</V>',
            ParseError, True
        ),
        (
            "Incorrectly encoded UTF-8 with 4-byte sequence "
            "with incorrect 3rd byte",
            b'<V>a\xF1\x80abc</V>',
            ParseError, True
        ),
        (
            "Incorrectly encoded UTF-8 with 4-byte sequence "
            "with incorrect 3rd byte that is an incorrect new start",
            b'<V>a\xF1\x80\xFFbc</V>',
            ParseError, True
        ),
        (
            "Incorrectly encoded UTF-8 with 4-byte sequence "
            "with incorrect 3rd byte that is a correct new start",
            b'<V>a\xF1\x80\xC2\x81c</V>',
            ParseError, True
        ),
    ]

    @pytest.mark.parametrize(
        "desc, utf8_string, exp_exc_type, condition",
        testcases
    )
    def test_check_invalid_utf8_sequences(
            self, desc, utf8_string, exp_exc_type, condition):
        # pylint: disable=no-self-use
        """
        Tests for tupletree.check_invalid_utf8_sequences()
        """

        if not condition:
            pytest.skip("Condition for test case not met")

        if exp_exc_type is None:

            unicode_string = tupletree.check_invalid_utf8_sequences(
                utf8_string, desc)

            assert isinstance(unicode_string, six.text_type), desc
            utf8_string_u = utf8_string.decode('utf-8')
            assert unicode_string == utf8_string_u, desc

        else:
            with pytest.raises(exp_exc_type):

                tupletree.check_invalid_utf8_sequences(utf8_string, desc)


class Test_check_invalid_xml_chars(object):
    # pylint: disable=too-few-public-methods
    """
    Tests for tupletree.check_invalid_xml_chars()
    """

    testcases = [
        # Each testcase has these items:
        # * desc: Short testcase description.
        # * xml_string: Input string for the function.
        # * exp_exc_type: Expected exception type, or None if expecting success.
        # * condition: Condition for testcase to run.

        # General cases
        (
            "Incorrect type of input string",
            b'<V>ab</V>',
            TypeError, True
        ),

        # Good cases
        (
            "Good case without special chars",
            u'<V>a</V>',
            None, True
        ),
        (
            "Good case with U+0009 (TAB), U+000A (NL), U+000D (CR) chars",
            u'<V>a\u0009b\u000Ac\u000Dd</V>',
            None, True
        ),
        (
            "Good case with U+0350 char",
            u'<V>a\u0350b</V>',
            None, True
        ),
        (
            "Good case with U+2013 char",
            u'<V>a\u2013b</V>',
            None, True
        ),
        (
            "Good case with U+10122 char",
            u'<V>a\u010122b</V>',
            None, True
        ),

        # Invalid XML characters
        (
            "Invalid XML char U+0008 (BEL)",
            u'<V>a\u0008b</V>',
            ParseError, True
        ),
        (
            "Invalid XML char U+0000",
            u'<V>a\u0000b</V>',
            ParseError, True
        ),
        (
            "Invalid XML chars U+0001 ... U+0008",
            u'<V>a\u0001b\u0002c\u0003d\u0004e\u0005f\u0006g\u0007h\u0008i</V>',
            ParseError, True
        ),
        (
            "Invalid XML chars U+000B ... U+000C",
            u'<V>a\u000Bb\u000Cc</V>',
            ParseError, True
        ),
        (
            "Invalid XML chars U+000E ... U+0015",
            u'<V>a\u000Eb\u000Fc\u0010d\u0011e\u0012f\u0013g\u0014h\u0015i</V>',
            ParseError, True
        ),
        (
            "Invalid XML chars U+0016 ... U+001D",
            u'<V>a\u0016b\u0017c\u0018d\u0019e\u001Af\u001Bg\u001Ch\u001Di</V>',
            ParseError, True
        ),
        (
            "Invalid XML chars U+001E ... U+001F",
            u'<V>a\u001Eb\u001Fc</V>',
            ParseError, True
        ),
    ]

    @pytest.mark.parametrize(
        "desc, xml_string, exp_exc_type, condition",
        testcases
    )
    def test_check_invalid_xml_chars(
            self, desc, xml_string, exp_exc_type, condition):
        # pylint: disable=no-self-use
        """
        Tests for tupletree.check_invalid_xml_chars()
        """

        if not condition:
            pytest.skip("Condition for test case not met")

        if exp_exc_type is None:

            tupletree.check_invalid_xml_chars(xml_string, desc)

        else:
            with pytest.raises(exp_exc_type):

                tupletree.check_invalid_xml_chars(xml_string, desc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Example of using Pull Operations to retrieve instances from a
    WBEM Server. This example allows either simplistic command line
    input with a fixed number of parameters or defaults to internal
    definitions if incorrect number of  cmd line arguments are supplied.

    It:
        Creates a connection
        Opens an enumeration session with OpenEnumerateInstances
        Executes a pull loop until the result.eos =True
        Displays overall statistics on the returns
    It also displays the results of the open and each pull in detail
"""

from __future__ import print_function

import sys
import datetime
from pywbem import WBEMConnection, CIMError, Error

# Default connection attributes. Used if not all arguments are
# supplied on the command line.
USERNAME = 'blah'
PASSWORD = 'blah'
TEST_CLASS = 'CIM_ComputerSystem'
TEST_NAMESPACE = 'root/cimv2'
SERVER_URL = 'http://localhost'

class ElapsedTimer(object):
    """
        Set up elapsed time timer. Calculates time between initiation
        and access.
    """
    def __init__(self):
        """ Initiate the object with current time"""
        self.start_time = datetime.datetime.now()

    def reset(self):
        """ Reset the start time for the timer"""
        self.start_time = datetime.datetime.now()

    def elapsed_ms(self):
        """ Get the elapsed time in milliseconds. returns floating
            point representation of elapsed time in seconds.
        """
        dt = datetime.datetime.now() - self.start_time
        return ((dt.days * 24 * 3600) + dt.seconds) * 1000  \
                + dt.microseconds / 1000.0

    def elapsed_sec(self):
        """ get the elapsed time in seconds. Returns floating
            point representation of time in seconds
        """
        return self.elapsed_ms() / 1000


def execute_request(conn, classname, max_open, max_pull):
    """
        Enumerate instances defined by the function's
        classname argument using the OpenEnumerateInstances and
        PullInstancesWithPath.

        * classname - Classname for the enumeration.

        * max_open - defines the maximum number of instances
          for the server to return for the open

        *max_pull defines the maximum number of instances for the
          WBEM server to return for each pull operation.

        Displays results of each open or pull operation including
        size, return parameters, and time to execute.

        Any exception exits the function.
    """
    start = ElapsedTimer()
    result = conn.OpenEnumerateInstances(classname,
                                         MaxObjectCount=max_open)


    print('open rtn eos=%s context=%s, count=%s time=%s ms' %
          (result.eos, result.context, len(result.instances),
           start.elapsed_ms()))

    # save instances since we reuse result
    insts = result.instances

    # loop to make pull requests until end_of_sequence received.
    pull_count = 0
    while not result.eos:
        pull_count += 1
        op_start = ElapsedTimer()
        result = conn.PullInstancesWithPath(result.context,
                                            MaxObjectCount=max_pull)

        insts.extend(result.instances)

        print('pull rtn eos=%s context=%s, insts=%s time=%s ms' %
              (result.eos, result.context, len(result.instances),
               op_start.elapsed_ms()))


    print('Result instance count=%s pull count=%s time=%.2f sec' % \
          (len(insts), pull_count, start.elapsed_sec()))
    return insts

def main():
    """
        Get arguments and call the execution function
    """

    # if less than required arguments, use the defaults
    if len(sys.argv) < 8:
        print("Usage: %s server_url username password namespace classname "
              "max_open, max_pull" %  sys.argv[0])
        server_url = SERVER_URL
        username = USERNAME
        password = PASSWORD
        namespace = TEST_NAMESPACE
        classname = TEST_CLASS
        max_open = 0
        max_pull = 100
    else:
        server_url = sys.argv[1]
        username = sys.argv[2]
        password = sys.argv[3]
        namespace = sys.argv[4]
        classname = sys.argv[5]
        max_open = sys.argv[6]
        max_pull = sys.argv[7]

    print('Parameters: server_url=%s\n username=%s\n namespace=%s\n' \
          ' classname=%s\n max_open=%s,\n max_pull=%s' % \
          (server_url, username, namespace, classname, max_open, max_pull))

    # connect to the server
    conn = WBEMConnection(server_url, (username, password),
                          default_namespace=namespace,
                          no_verification=True)

    #Call method to execute the enumeration sequence and return instances
    try:
        instances = execute_request(conn, classname, max_open, max_pull)

        # print the resulting instances
        for instance in instances:
            print('\npath=%s\n%s' % (instance.path, instance.tomof()))

    # handle exceptions
    except CIMError as ce:
        print('Operation Failed: CIMError: code=%s, Description=%s' % \
              (ce.status_code_name, ce.status_description))
        sys.exit(1)
    except Error as err:
        print ("Operation failed: %s" % err)
        sys.exit(1)

    return 0

if __name__ == '__main__':
    sys.exit(main())






"""
Wbemcli scriptlet that removes all subscriptions, filters, and destination
objects from the defined wbem connection
"""

from pywbem import WBEMServer


def display_path_info(descriptor, class_, namespace):
    """
    Display info on the instance names of the class parameter retrieved
    from the interop namespace of the server defined by CONN.

    The descriptor is text defining the class(ex. Filter)).
    Returns a list of the paths
    """
    paths = CONN.EnumerateInstanceNames(class_, namespace=interop)  # noqa: F821
    print('%ss: count=%s' % (descriptor, len(paths)))
    for path in paths:
        print('%s: %s' % (descriptor, path))
    return paths


def delete_paths(paths):
    """Delete all the CIMInstanceNames in paths paramter from the server."""
    for path in paths:
        CONN.DeleteInstance(path)  # noqa: F821


MY_SERVER = WBEMServer(CONN)  # noqa: F821
interop = MY_SERVER.interop_ns
print('Interop namespace=%s' % interop)

filter_paths = display_path_info("Filter", 'CIM_IndicationFilter', interop)
dest_paths = display_path_info("Destination", 'CIM_ListenerDestination',
                               interop)
sub_paths = display_path_info("Subscription", 'CIM_IndicationSubscription',
                              interop)

# If any paths exist, remove them and then redisplay the results
if filter_paths or dest_paths or sub_paths:

    delete_paths(sub_paths)
    delete_paths(dest_paths)
    delete_paths(filter_paths)

    filter_paths = display_path_info("Filter", 'CIM_IndicationFilter', interop)
    dest_paths = display_path_info("Destination", 'CIM_ListenerDestination',
                                   interop)
    sub_paths = display_path_info("Subscription", 'CIM_IndicationSubscription',
                                  interop)

print('Quit Scriplet. Quiting wbemcli')
quit()



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
The :class:`~pywbem.ValueMapping` class supports translating between the values
of an integer-typed CIM element (e.g. property, method, or parameter) that is
qualified with the `ValueMap` and `Values` qualifiers, and the corresponding
values of the `Values` qualifier, in both directions.

This class supports value ranges (e.g. ``"4..6"``) and the unclaimed marker
(``".."``).
"""

import re
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
import six

from .cim_types import CIMInt, type_from_name
from .cim_obj import CIMProperty, CIMMethod, CIMParameter
from ._utils import _format, _integerValue_to_int

__all__ = ['ValueMapping']


class ValueMapping(object):
    """
    *New in pywbem 0.9 as experimental and finalized in 0.10.*

    A utility class that supports translating between the values of an
    integer-typed CIM element (property, method, parameter) that is qualified
    with the `ValueMap` and `Values` qualifiers, and the corresponding values
    of the `Values` qualifier, in both directions.

    This is done by retrieving the CIM class definition defining the CIM
    element in question, and by inspecting its `ValueMap` and `Values`
    qualifiers.

    The translation is performed by the
    :meth:`~pywbem.ValueMapping.tovalues` and
    :meth:`~pywbem.ValueMapping.tobinary` methods.

    Instances of this class must be created through one of the factory class
    methods: :meth:`~pywbem.ValueMapping.for_property`,
    :meth:`~pywbem.ValueMapping.for_method`, or
    :meth:`~pywbem.ValueMapping.for_parameter`.

    Value ranges (``"2..4"``) and the indicator for unclaimed values (``".."``)
    in the `ValueMap` qualifier are supported.

    All representations of the integer values in the `ValueMap` qualifier are
    supported (decimal, binary, octal, hexadecimal), consistent with the
    definition of the `ValueMap` qualifier in :term:`DSP0004`.

    Example:

      Given the following definition of a property in MOF:

      .. code-block:: text

          class CIM_Foo {

                [ValueMap{ "0", "2..4", "..6", "7..", "9", ".." },
                 Values{ "zero", "two-four", "five-six", "seven-eight", "nine",
                 "unclaimed"}]
             uint16 MyProp;

          };

      Assuming this class exists in a WBEM server, the following code will
      get the class from the server, create a value mapping for this property,
      and look up the `Values` strings that correspond to binary property
      values. This is useful when preparing binary property values for human
      consumption::

          namespace = 'root/cimv2'
          conn = pywbem.WBEMConnection(...)  # WBEM server

          myprop_vm = pywbem.ValueMapping.for_property(
              conn, namespace, 'CIM_Foo', 'MyProp')

          print("Binary value: Values string")
          for bin_value in range(0, 12):
              values_str = myprop_vm.tovalues(bin_value)
              print("{0:12}: {1!r}".format(bin_value, values_str))

      Resulting output:

      .. code-block:: text

          Binary value: Values string
                     0: 'zero'
                     1: 'unclaimed'
                     2: 'two-four'
                     3: 'two-four'
                     4: 'two-four'
                     5: 'five-six'
                     6: 'five-six'
                     7: 'seven-eight'
                     8: 'seven-eight'
                     9: 'nine'
                    10: 'unclaimed'
                    11: 'unclaimed'

      Translating in the other direction is also of interest, for example when
      processing values that are provided by humans in terms of the `Values`
      strings, or when using the pywbem mock support and the test cases specify
      property values for CIM instances in terms of the more human-readable
      `Values` strings.

      Again, assuming the class shown above exists in a WBEM server, the
      following code will get the class from the server, create a value
      mapping for this property, and look up the binary property values from
      the `Values` strings::

          namespace = 'root/cimv2'
          conn = pywbem.WBEMConnection(...)  # WBEM server

          myprop_vm = pywbem.ValueMapping.for_property(
              conn, namespace, 'CIM_Foo', 'MyProp')

          values_strs = ["zero", "two-four", "five-six", "seven-eight", "nine",
                         "unclaimed"]

          print("Values string: Binary value")
          for values_str in values_strs:
              bin_value = myprop_vm.tobinary(values_str)
              print("{0:12}: {1!r}".format(values_str, bin_value))

      Resulting output:

      .. code-block:: text

          Values string: Binary value
                 'zero': 0
             'two-four': (2, 4)
             'five-six': (5, 6)
          'seven-eight': (7, 8)
                 'nine': 9
            'unclaimed': None

      Iterating through the pairs of `ValueMap` and `Values` entries is
      also possible. Assuming the class shown above exists in a WBEM server, the
      following code will get the class from the server, and iterate through the
      value mapping::

          namespace = 'root/cimv2'
          conn = pywbem.WBEMConnection(...)  # WBEM server

          myprop_vm = pywbem.ValueMapping.for_property(
              conn, namespace, 'CIM_Foo', 'MyProp')

          print("Values string: Binary value")
          for bin_value, values_str in myprop_vm.items():
              print("{0:12}: {1!r}".format(values_str, bin_value))

      Resulting output:

      .. code-block:: text

          Values string: Binary value
                 'zero': 0
             'two-four': (2, 4)
             'five-six': (5, 6)
          'seven-eight': (7, 8)
                 'nine': 9
            'unclaimed': None
    """

    def __init__(self):

        self._conn = None
        self._namespace = None
        self._classname = None
        self._propname = None
        self._methodname = None
        self._parametername = None

        self._element_obj = None

        # Attributes for converting binary values to Values strings:
        self._b2v_single_dict = {}  # for single values; bin: values
        self._b2v_range_tuple_list = []  # for value ranges; tuple(lo,hi,values)
        self._b2v_unclaimed = None  # value of the unclaimed indicator '..'

        # Attributes for converting Values strings to binary values:
        self._v2b_dict = {}  # values: bin (int or tuple)

    @classmethod
    def for_property(cls, server, namespace, classname, propname):
        # pylint: disable=line-too-long
        """
        Factory method that returns a new :class:`~pywbem.ValueMapping`
        instance that maps CIM property values to the `Values` qualifier
        defined on that property.

        If a `Values` qualifier is defined but no `ValueMap` qualifier, a
        default of 0-based consecutive numbers is applied (that is the default
        defined in :term:`DSP0004`).

        Parameters:

          server (:class:`~pywbem.WBEMConnection` or :class:`~pywbem.WBEMServer`):
            The connection to the WBEM server containing the namespace.

          namespace (:term:`string`):
            Name of the CIM namespace containing the class.
            If `None`, the default namespace of the connection will be used.

          classname (:term:`string`):
            Name of the CIM class exposing the property. The property can be
            defined in that class or inherited into that class.

          propname (:term:`string`):
            Name of the CIM property that defines the `Values` / `ValueMap`
            qualifiers.

        Returns:

            The new :class:`~pywbem.ValueMapping` instance.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            KeyError: The CIM property does not exist in the CIM class.
            TypeError: The CIM property is not integer-typed.
            ValueError: No `Values` qualifier defined on the CIM property.
            ValueError: Invalid integer representation in `ValueMap` qualifier
              defined on the CIM property.
        """  # noqa: E501

        conn = server
        try:
            get_class = conn.GetClass
        except AttributeError:
            conn = server.conn
            get_class = conn.GetClass

        class_obj = get_class(ClassName=classname,
                              namespace=namespace,
                              LocalOnly=False,
                              IncludeQualifiers=True)
        try:
            property_obj = class_obj.properties[propname]
        except KeyError:
            raise KeyError(
                _format("Class {0!A} (in {1!A}) does not have a property "
                        "{2!A}", classname, namespace, propname))

        new_vm = cls._create_for_element(property_obj, conn, namespace,
                                         classname, propname=propname)

        return new_vm

    @classmethod
    def for_method(cls, server, namespace, classname, methodname):
        # pylint: disable=line-too-long
        """
        Factory method that returns a new :class:`~pywbem.ValueMapping`
        instance that maps CIM method return values to the `Values` qualifier
        of that method.

        If a `Values` qualifier is defined but no `ValueMap` qualifier, a
        default of 0-based consecutive numbers is applied (that is the default
        defined in :term:`DSP0004`).

        Parameters:

          server (:class:`~pywbem.WBEMConnection` or :class:`~pywbem.WBEMServer`):
            The connection to the WBEM server containing the namespace.

          namespace (:term:`string`):
            Name of the CIM namespace containing the class.

          classname (:term:`string`):
            Name of the CIM class exposing the method. The method can be
            defined in that class or inherited into that class.

          methodname (:term:`string`):
            Name of the CIM method that defines the `Values` / `ValueMap`
            qualifiers.

        Returns:

            The new :class:`~pywbem.ValueMapping` instance.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            KeyError: The CIM method does not exist in the CIM class.
            TypeError: The CIM method is not integer-typed.
            ValueError: No `Values` qualifier defined on the CIM method.
            ValueError: Invalid integer representation in `ValueMap` qualifier
              defined on the CIM method.
        """  # noqa: E501

        conn = server
        try:
            get_class = conn.GetClass
        except AttributeError:
            conn = server.conn
            get_class = conn.GetClass

        class_obj = get_class(ClassName=classname,
                              namespace=namespace,
                              LocalOnly=False,
                              IncludeQualifiers=True)
        try:
            method_obj = class_obj.methods[methodname]
        except KeyError:
            raise KeyError(
                _format("Class {0!A} (in {1!A}) does not have a method {2!A}",
                        classname, namespace, methodname))

        new_vm = cls._create_for_element(method_obj, conn, namespace,
                                         classname, methodname=methodname)

        return new_vm

    @classmethod
    def for_parameter(cls, server, namespace, classname, methodname,
                      parametername):
        # pylint: disable=line-too-long
        """
        Factory method that returns a new :class:`~pywbem.ValueMapping`
        instance that maps CIM parameter values to the `Values` qualifier
        defined on that parameter.

        If a `Values` qualifier is defined but no `ValueMap` qualifier, a
        default of 0-based consecutive numbers is applied (that is the default
        defined in :term:`DSP0004`).

        Parameters:

          server (:class:`~pywbem.WBEMConnection` or :class:`~pywbem.WBEMServer`):
            The connection to the WBEM server containing the namespace.

          namespace (:term:`string`):
            Name of the CIM namespace containing the class.

          classname (:term:`string`):
            Name of the CIM class exposing the method. The method can be
            defined in that class or inherited into that class.

          methodname (:term:`string`):
            Name of the CIM method that has the parameter.

          parametername (:term:`string`):
            Name of the CIM parameter that defines the `Values` / `ValueMap`
            qualifiers.

        Returns:

            The new :class:`~pywbem.ValueMapping` instance.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            KeyError: The CIM method does not exist in the CIM class.
            KeyError: The CIM parameter does not exist in the CIM method.
            TypeError: The CIM parameter is not integer-typed.
            ValueError: No `Values` qualifier defined on the CIM parameter.
            ValueError: Invalid integer representation in `ValueMap` qualifier
              defined on the CIM parameter.
        """  # noqa: E501

        conn = server
        try:
            get_class = conn.GetClass
        except AttributeError:
            conn = server.conn
            get_class = conn.GetClass

        class_obj = get_class(ClassName=classname,
                              namespace=namespace,
                              LocalOnly=False,
                              IncludeQualifiers=True)
        try:
            method_obj = class_obj.methods[methodname]
        except KeyError:
            raise KeyError(
                _format("Class {0!A} (in {1!A}) does not have a method {2!A}",
                        classname, namespace, methodname))
        try:
            parameter_obj = method_obj.parameters[parametername]
        except KeyError:
            raise KeyError(
                _format("Method {0!A} in class {1!A} (in {2!A}) does not have "
                        "a parameter {3!A}", methodname, classname, namespace,
                        parametername))

        new_vm = cls._create_for_element(parameter_obj, conn, namespace,
                                         classname, methodname=methodname,
                                         parametername=parametername)

        return new_vm

    def _values_tuple(self, i, valuemap_list, values_list, cimtype):
        """
        Return a tuple for the value range or unclaimed marker at position i,
        with these items:

        * lo - low value of the range
        * hi - high value of the range (can be equal to lo)
        * values - value of `Values` qualifier for this position

        Parameters:

          i (integer): position into valuemap_list and values_list

          valuemap_list (list of strings): `ValueMap` qualifier value

          values_list (list of strings): `Values` qualifier value

          cimtype (type): CIM type of the CIM element

        Raises:

            ValueError: Invalid integer representation in `ValueMap` qualifier
              defined on the CIM element.
        """
        values_str = values_list[i]
        valuemap_str = valuemap_list[i]
        m = re.match(r'^(.*)\.\.(.*)$', valuemap_str)
        if m is None:
            valuemap_int = self._to_int(valuemap_str)
            return (valuemap_int, valuemap_int, values_str)

        # match found
        lo = m.group(1)
        if lo == '':
            if i == 0:
                lo = cimtype.minvalue
            else:
                _, previous_hi, _ = self._values_tuple(
                    i - 1, valuemap_list, values_list, cimtype)
                lo = previous_hi + 1
        else:
            lo = self._to_int(lo)

        hi = m.group(2)
        if hi == '':
            if i == len(valuemap_list) - 1:
                hi = cimtype.maxvalue
            else:
                next_lo, _, _ = self._values_tuple(
                    i + 1, valuemap_list, values_list, cimtype)
                hi = next_lo - 1
        else:
            hi = self._to_int(hi)
        return (lo, hi, values_str)

    def _to_int(self, val_str):
        """Conver val_str to an integer or raise ValueError"""
        val = _integerValue_to_int(val_str)
        if val is None:
            raise ValueError(
                _format("The value-mapped {0} has an invalid integer "
                        "representation in a ValueMap entry: {1!A}",
                        self._element_str(), val_str))
        return val

    @classmethod
    def _create_for_element(cls, element_obj, conn, namespace, classname,
                            propname=None, methodname=None, parametername=None):
        # pylint: disable=line-too-long
        """
        Return a new :class:`~pywbem.ValueMapping` instance for the specified
        CIM element.

        If a `Values` qualifier is defined but no `ValueMap` qualifier, a
        default of 0-based consecutive numbers is applied (that is the default
        defined in :term:`DSP0004`).

        Parameters:

          element_obj (:class:`~pywbem.CIMProperty`, :class:`~pywbem.CIMMethod`, or :class:`~pywbem.CIMParameter`):
            The CIM element on which the qualifiers are defined.

          conn (:class:`~pywbem.WBEMConnection`):
            The connection to the WBEM server containing the namespace.

          namespace (:term:`string`):
            Name of the CIM namespace containing the class.

          classname (:term:`string`):
            Name of the CIM class exposing the method. The method can be
            defined in that class or inherited into that class.

          propname (:term:`string`):
            Name of the CIM property that defines the `Values` / `ValueMap`
            qualifiers.

          methodname (:term:`string`):
            Name of the CIM method that has the parameter.

          parametername (:term:`string`):
            Name of the CIM parameter that defines the `Values` / `ValueMap`
            qualifiers.

        Returns:

            The created :class:`~pywbem.ValueMapping` instance for the specified
            CIM element.

        Raises:

            TypeError: The CIM element is not integer-typed.
            ValueError: No `Values` qualifier defined on the CIM element.
            ValueError: Invalid integer representation in `ValueMap` qualifier
              defined on the CIM element.
        """  # noqa: E501
        # pylint: enable=line-too-long

        # pylint: disable=protected-access

        vm = ValueMapping()
        vm._element_obj = element_obj
        vm._conn = conn
        vm._namespace = namespace
        vm._classname = classname
        vm._propname = propname
        vm._methodname = methodname
        vm._parametername = parametername

        try:
            typename = element_obj.type  # Property, Parameter
        except AttributeError:
            typename = element_obj.return_type  # Method

        cimtype = type_from_name(typename)

        if not issubclass(cimtype, CIMInt):
            raise TypeError(
                _format("The value-mapped {0} is not integer-typed, but "
                        "has CIM type: {1}", vm._element_str(), typename))

        values_qual = element_obj.qualifiers.get('Values', None)
        if values_qual is None:
            # DSP0004 defines no default for a missing Values qualifier
            raise ValueError(
                _format("The value-mapped {0} has no Values qualifier "
                        "defined", vm._element_str()))
        values_list = values_qual.value

        valuemap_qual = element_obj.qualifiers.get('ValueMap', None)
        if valuemap_qual is None:
            # DSP0004 defines a default of consecutive index numbers
            vm._b2v_single_dict = dict(zip(range(0, len(values_list)),
                                           values_list))
            vm._b2v_range_tuple_list = []
            vm._b2v_unclaimed = None
            vm._v2b_dict = OrderedDict(zip(values_list,
                                           range(0, len(values_list))))
        else:
            vm._b2v_single_dict = {}
            vm._b2v_range_tuple_list = []
            vm._b2v_unclaimed = None
            vm._v2b_dict = OrderedDict()
            valuemap_list = valuemap_qual.value
            for i, valuemap_str in enumerate(valuemap_list):
                values_str = values_list[i]
                if valuemap_str == '..':
                    vm._b2v_unclaimed = values_str
                    vm._v2b_dict[values_str] = None
                else:
                    lo, hi, values_str = vm._values_tuple(
                        i, valuemap_list, values_list, cimtype)
                    if lo == hi:
                        # single value
                        vm._b2v_single_dict[lo] = values_str
                        vm._v2b_dict[values_str] = lo
                    else:
                        # value range
                        vm._b2v_range_tuple_list.append((lo, hi, values_str))
                        vm._v2b_dict[values_str] = (lo, hi)

        return vm

    def _element_str(self):
        """
        Return a string that identifies the value-mapped element.
        """
        # pylint: disable=no-else-return
        if isinstance(self.element, CIMProperty):
            return _format("property {0!A} in class {1!A} (in {2!A})",
                           self.propname, self.classname, self.namespace)
        elif isinstance(self.element, CIMMethod):
            return _format("method {0!A} in class {1!A} (in {2!A})",
                           self.methodname, self.classname, self.namespace)
        assert isinstance(self.element, CIMParameter)
        return _format("parameter {0!A} of method {1!A} in class {2!A} "
                       "(in {3!A})",
                       self.parametername, self.methodname, self.classname,
                       self.namespace)

    def __repr__(self):
        """
        Return a representation of the :class:`~pywbem.ValueMapping` object
        with all attributes, that is suitable for debugging.
        """
        return _format(
            "ValueMapping("
            "_conn={s._conn!A}, "
            "_namespace={s._namespace!A}, "
            "_classname={s._classname!A}, "
            "_propname={s._propname!A}, "
            "_methodname={s._methodname!A}, "
            "_parametername={s._parametername!A}, "
            "_element_obj={s._element_obj!A}, "
            "_b2v_single_dict={s._b2v_single_dict!A}, "
            "_b2v_range_tuple_list={s._b2v_range_tuple_list!A}, "
            "_b2v_unclaimed={s._b2v_unclaimed!A}, "
            "_v2b_dict={s._v2b_dict!A})",
            s=self)

    @property
    def conn(self):
        """
        :class:`~pywbem.WBEMConnection`:
        Connection to the WBEM server containing the CIM namespace (that
        contains the mapped CIM element).
        """
        return self._conn

    @property
    def namespace(self):
        """
        :term:`string`: Name of the CIM namespace containing the class that
        defines the mapped CIM element.
        """
        return self._namespace

    @property
    def classname(self):
        """
        :term:`string`: Name of the CIM class defining the mapped CIM element.
        """
        return self._classname

    @property
    def propname(self):
        """
        :term:`string`: Name of the CIM property that is mapped. `None`, if
        no property is mapped.
        """
        return self._propname

    @property
    def methodname(self):
        """
        :term:`string`: Name of the CIM method, that either is mapped itself,
        or that has the parameter that is mapped.  `None`, if no method or
        parameter is mapped.
        """
        return self._methodname

    @property
    def parametername(self):
        """
        :term:`string`: Name of the CIM parameter that is mapped. `None`, if
        no parameter is mapped.
        """
        return self._parametername

    @property
    def element(self):
        # pylint: disable=line-too-long
        """
        :class:`~pywbem.CIMProperty`, :class:`~pywbem.CIMMethod`, or :class:`~pywbem.CIMParameter`:
        The mapped CIM element.
        """  # noqa: E501
        return self._element_obj

    def tovalues(self, element_value):
        """
        Return the `Values` string for an element value, based upon this value
        mapping.

        Parameters:

          element_value (:term:`integer` or :class:`~pywbem.CIMInt`):
            The value of the CIM element (property, method, parameter).

        Returns:

          :term:`string`:
            The `Values` string for the element value.

        Raises:

          ValueError: Element value outside of the set defined by `ValueMap`.
          TypeError: Element value is not an integer type.
        """

        if not isinstance(element_value, (six.integer_types, CIMInt)):
            raise TypeError(
                _format("The value for value-mapped {0} is not "
                        "integer-typed, but has Python type: {1}",
                        self._element_str(), type(element_value)))

        # try single value
        try:
            return self._b2v_single_dict[element_value]
        except KeyError:
            pass

        # try value ranges
        for range_tuple in self._b2v_range_tuple_list:
            lo, hi, values_str = range_tuple
            if lo <= element_value <= hi:
                return values_str

        # try catch-all '..'
        if self._b2v_unclaimed is not None:
            return self._b2v_unclaimed

        raise ValueError(
            _format("The value for value-mapped {0} is outside of the set "
                    "defined by its ValueMap qualifier: {1!A}",
                    self._element_str(), element_value))

    def tobinary(self, values_str):
        """
        Return the integer value or values for a `Values` string, based upon
        this value mapping.

        Any returned integer value is represented as the CIM type of the
        element (e.g. :class:`~pywbem.Uint16`).

        If the `Values` string corresponds to a single value, that single value
        is returned as an integer.

        If the `Values` string corresponds to a value range (e.g. "1.." or
        "..2" or "3..4"), that value range is returned as a tuple with two items
        that are the lowest and the highest value of the range. That is the
        case also when the value range is open on the left side or right side.

        If the `Values` string corresponds to the `unclaimed` indicator "..",
        `None` is returned.

        Parameters:

          values_str (:term:`string`):
            The `Values` string for the element value.

        Returns:

          :class:`~pywbem.CIMInt` or tuple of :class:`~pywbem.CIMInt` or `None`:
            The element value or value range corresponding to the `Values`
            string, or `None` for unclaimed.

        Raises:

          ValueError: `Values` string outside of the set defined by `Values`.
          TypeError: `Values` string is not a string type.
        """

        if not isinstance(values_str, six.string_types):
            raise TypeError(
                _format("The values string for value-mapped {0} is not "
                        "string-typed, but has Python type: {1}",
                        self._element_str(), type(values_str)))

        try:
            return self._v2b_dict[values_str]
        except KeyError:
            raise ValueError(
                _format("The values string for value-mapped {0} is outside "
                        "of the set defined by its Values qualifier: {1!A}",
                        self._element_str(), values_str))

    def items(self):
        """
        Generator that iterates through the items of the value mapping. The
        items are the array entries of the `Values` and `ValueMap` qualifiers,
        and they are iterated in the order specified in the arrays.
        If the `ValueMap` qualifier is not specified, the default of consecutive
        integers starting at 0 is used as a default, consistent with
        :term:`DSP0004`.

        Each iterated item is a tuple of integer value(s) representing the
        `ValueMap` array entry, and the corresponding `Values` string.
        Any integer value in the iterated items is represented as the CIM type
        of the element (e.g. :class:`~pywbem.Uint16`).

        If the `Values` string corresponds to a single element value, the
        first tuple item is that single integer value.

        If the `Values` string corresponds to a value range (e.g. "1.." or
        "..2" or "3..4"), that value range is returned as a tuple with two items
        that are the lowest and the highest value of the range. That is the
        case also when the value range is open on the left side or right side.

        If the `Values` string corresponds to the `unclaimed` indicator "..",
        the first tuple item is `None`.

        Returns:

          :term:`iterator` for tuple of integer value(s) and `Values` string.
        """

        for values_str in self._v2b_dict:
            element_value = self._v2b_dict[values_str]
            yield element_value, values_str



#
# (C) Copyright 2018 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
Test for the WBEMServer class  in pywbem._server.py that uses the pywbem_mock
support package the methods of the class. Mock is required since, testing
WBEMServe requires access to a WBEMServer.
"""
from __future__ import absolute_import, print_function

import os
import re
from socket import getfqdn

from pywbem import WBEMServer, CIMClassName, WBEMSubscriptionManager, \
    CIMInstance

from pywbem._subscription_manager import SUBSCRIPTION_CLASSNAME, \
    DESTINATION_CLASSNAME, FILTER_CLASSNAME

from ..utils.dmtf_mof_schema_def import DMTF_TEST_SCHEMA_VER
from ..utils.wbemserver_mock import WbemServerMock


# Location of DMTF schema directory used by all tests.
# This directory is permanent and should not be removed.
TESTSUITE_SCHEMA_DIR = os.path.join('tests', 'schema')

VERBOSE = False


class BaseMethodsForTests(object):
    """
    Base Class for tests contains all common methods and class level
    setup.
    """

    def setup_method(self):
        """
        Create the Fake connection and setup the connection. This
        method first initialized the WbemServerMock and then suplements
        it with the classes required for the subscription manager.
        """
        server = WbemServerMock(interop_ns='interop')
        # pylint: disable=attribute-defined-outside-init
        self.conn = server.wbem_server.conn
        classnames = ['CIM_IndicationSubscription',
                      'CIM_ListenerDestinationCIMXML',
                      'CIM_IndicationFilter', ]

        self.conn.compile_dmtf_schema(DMTF_TEST_SCHEMA_VER,
                                      TESTSUITE_SCHEMA_DIR,
                                      class_names=classnames, verbose=False)

    @staticmethod
    def inst_from_classname(conn, class_name, namespace=None,
                            property_list=None,
                            property_values=None,
                            include_missing_properties=True,
                            include_path=True):
        # TODO AM 8/18 The inst_from_classname() method is not used.
        """
        Build instance from class using class_name property to get class
        from a repository.
        """
        cls = conn.GetClass(class_name, namespace=namespace, LocalOnly=False,
                            IncludeQualifiers=True, include_class_origin=True,
                            property_list=property_list)

        return CIMInstance.from_class(
            cls, namespace=namespace, property_values=property_values,
            include_missing_properties=include_missing_properties,
            include_path=include_path)

    def add_filter(self, sub_mgr, server_id, filter_id, owned=True):
        """
        Create a single filter definition in the sub_mgr (which adds it to
        the repository) and returns the path of the new filter instance.
        This creates a filter specifically these tests
        """
        # pylint: disable=attribute-defined-outside-init
        self.test_class = 'Test_IndicationProviderClass'
        self.test_class_namespace = 'test/TestProvider'

        self.test_query = 'SELECT * from %s' % self.test_class
        self.test_classname = CIMClassName(self.test_class,
                                           namespace=self.test_class_namespace)

        filter_ = sub_mgr.add_filter(server_id,
                                     self.test_class_namespace,
                                     self.test_query,
                                     query_language="DMTF:CQL",
                                     filter_id=filter_id,
                                     owned=owned)
        return filter_.path

    def get_objects_from_server(self):
        """
        Using Server class, get count of Filters, Subscriptions, Destinations
        from server as confirmation outside of SubscriptionManagerCode.
        """
        this_host = getfqdn()

        server = WBEMServer(self.conn)

        dest_name_pattern = re.compile(r'^pywbemdestination:')
        dest_insts = server.conn.EnumerateInstances(
            DESTINATION_CLASSNAME, namespace=server.interop_ns)
        valid_dests = []
        for inst in dest_insts:
            if re.match(dest_name_pattern, inst.path.keybindings['Name']) \
                    and inst.path.keybindings['SystemName'] == this_host:
                valid_dests.append(inst)

        filter_instances = server.conn.EnumerateInstances(
            FILTER_CLASSNAME, namespace=server.interop_ns)
        valid_filters = []
        filter_name_pattern = re.compile(r'^pywbemfilter')
        for inst in filter_instances:
            if re.match(filter_name_pattern, inst.path.keybindings['Name']) \
                    and inst.path.keybindings['SystemName'] == this_host:
                valid_filters.append(inst)

        filter_paths = [inst.path for inst in valid_filters]
        destination_paths = [inst.path for inst in valid_dests]

        sub_insts = server.conn.EnumerateInstances(
            SUBSCRIPTION_CLASSNAME, namespace=server.interop_ns)
        valid_subscriptions = []
        for inst in sub_insts:
            if inst.path.keybindings['Filter'] in filter_paths \
                    or inst.path.keybindings['Handler'] in destination_paths:
                valid_subscriptions.append(inst)

        if VERBOSE:
            print('All objects: filters=%s dests=%s subs=%s' %
                  (len(filter_instances),
                   len(dest_insts), len(sub_insts)))
            print('Pywbem objects: filters=%s dests=%s subs=%s' %
                  (len(filter_paths), len(destination_paths),
                   len(valid_subscriptions)))

        return sum([len(filter_paths), len(destination_paths),
                    len(valid_subscriptions)])

    def confirm_created(self, sub_mgr, server_id, filter_path,
                        subscription_paths, owned=True):
        # pylint: disable=no-self-use
        """
        Confirm that filters and subscription paths provided with call
        exist in the appropriated lists.
        """
        if owned:
            owned_filters = sub_mgr.get_owned_filters(server_id)
            owned_filter_paths = [inst.path for inst in owned_filters]

            assert filter_path in owned_filter_paths

            owned_subscriptions = sub_mgr.get_owned_subscriptions(server_id)
            owned_sub_paths = [inst.path for inst in owned_subscriptions]
            for subscription_path in owned_sub_paths:
                assert subscription_path in owned_sub_paths

        all_filters = sub_mgr.get_all_filters(server_id)
        all_filter_paths = [inst.path for inst in all_filters]
        assert filter_path in all_filter_paths

        all_subscriptions = sub_mgr.get_all_subscriptions(server_id)
        all_sub_paths = [inst.path for inst in all_subscriptions]
        for subscription_path in subscription_paths:
            assert subscription_path in all_sub_paths

    def confirm_removed(self, sub_mgr, server_id, filter_paths,
                        subscription_paths):
        # pylint: disable=no-self-use
        """
        When filter_path and subscription path are removed, this
        confirms that results are correct both in the local subscription
        manager and the remote WBEM server.
        """
        if not isinstance(filter_paths, list):
            filter_paths = [filter_paths]

        owned_filters = sub_mgr.get_owned_filters(server_id)
        for filter_path in filter_paths:
            assert filter_path not in owned_filters

        # confirm not in owned subscriptions
        if not isinstance(subscription_paths, list):
            subscription_paths = [subscription_paths]

        owned_subscriptions = sub_mgr.get_owned_subscriptions(server_id)
        owned_sub_paths = [inst.path for inst in owned_subscriptions]
        for subscription_path in subscription_paths:
            assert subscription_path not in owned_sub_paths

        all_filters = sub_mgr.get_all_filters(server_id)
        all_filter_paths = [inst.path for inst in all_filters]
        for filter_path in filter_paths:
            assert filter_path not in all_filter_paths

        all_subscriptions = sub_mgr.get_all_subscriptions(server_id)
        all_sub_paths = [inst.path for inst in all_subscriptions]
        for subscription_path in subscription_paths:
            assert subscription_path not in all_sub_paths

    def get_object_count(self, sub_mgr, server_id):
        """
        Return count of all filters, subscriptions, and dests
        for the defined sub_mgr and server_id. Accumulates the filter,
        subscription, dests that should represent what is in the server
        and local.
        """
        # pylint: disable=no-self-use
        return sum(self.count_outstanding(sub_mgr, server_id))

    def empty_expected(self, sub_mgr, server_id):
        """ TODO """
        # pylint: disable=no-self-use

        counts = self.count_outstanding(sub_mgr, server_id)
        if sum(counts) == 0:
            return True

        print('ERROR: Server_id=%s. Unreleased filters=%s, subs=%s, dest=%s' %
              (server_id, counts[0], counts[1], counts[2]))
        return False

    def count_outstanding(self, sub_mgr, server_id):
        """
            Count outstanding filters, subscriptions, and destinations and
            return tuple with the counts
        """
        # pylint: disable=no-self-use

        filters = len(sub_mgr.get_all_filters(server_id))
        subscriptions = len(sub_mgr.get_all_subscriptions(server_id))
        dests = len(sub_mgr.get_all_subscriptions(server_id))

        return (filters, subscriptions, dests)


class TestSubMgrClass(BaseMethodsForTests):
    """Test of subscription manager"""

    def test_create_owned_subscription(self):
        """
        Test Basic Creation of sub mgr, creation of owned subscription
        and cleanup based on the WBEMSubscriptionManager context manager.
        """

        sm = "test_create_delete_subscription"
        server = WBEMServer(self.conn)

        listener_url = '%s:%s' % (self.conn.url, 50000)

        with WBEMSubscriptionManager(subscription_manager_id=sm) as sub_mgr:

            server_id = sub_mgr.add_server(server)
            sub_mgr.add_listener_destinations(server_id, listener_url,
                                              owned=True)
            filter_path = self.add_filter(sub_mgr, server_id, 'NotUsed',
                                          owned=True)
            subscriptions = sub_mgr.add_subscriptions(server_id,
                                                      filter_path,
                                                      owned=True)
            subscription_paths = [inst.path for inst in subscriptions]

            # self.conn.display_repository()

            self.confirm_created(sub_mgr, server_id, filter_path,
                                 subscription_paths, owned=True)

            # confirm destination instance paths match
            assert sub_mgr.get_all_destinations(server_id)
            # TODO: ks Finish this test completely when we add other
            #  changes for filter ids

            sub_mgr.remove_subscriptions(server_id, subscription_paths)
            sub_mgr.remove_filter(server_id, filter_path)

            self.confirm_removed(sub_mgr, server_id, filter_path,
                                 subscription_paths)

            assert self.get_object_count(sub_mgr, server_id) == 0

            sub_mgr.remove_server(server_id)

        # confirm no filters, destinations, subscriptions in server
        assert self.get_objects_from_server() == 0

    def test_create_not_owned_subscription(self):
        """
        Test creating not_owned filter, destination and filter and determining
        if they are retained by the server after the subscription manager
        is closed.

        """
        sm = "test_create_delete_subscription"
        server = WBEMServer(self.conn)

        listener_url = '%s:%s' % (self.conn.url, 50000)

        # Create a single not_owned subscription
        with WBEMSubscriptionManager(subscription_manager_id=sm) as sub_mgr:
            server_id = sub_mgr.add_server(server)
            dests = sub_mgr.add_listener_destinations(server_id, listener_url,
                                                      owned=False)
            dest_paths = [dest.path for dest in dests]
            filter_path = self.add_filter(sub_mgr, server_id, 'NotUsed',
                                          owned=False)
            subscriptions = sub_mgr.add_subscriptions(
                server_id, filter_path,
                destination_paths=dest_paths,
                owned=False)
            subscription_paths = [inst.path for inst in subscriptions]

            self.confirm_created(sub_mgr, server_id, filter_path,
                                 subscription_paths, owned=False)
            assert self.get_object_count(sub_mgr, server_id)

        # Test that subscriptions instances are still in repo
        # self.conn.display_repository()
        assert self.get_objects_from_server() == 3

        # Create a new submgr and and test for filters, etc. retrieved.
        with WBEMSubscriptionManager(subscription_manager_id=sm) as sub_mgr:
            server_id = sub_mgr.add_server(server)
            assert len(sub_mgr.get_all_destinations(server_id)) == 1
            assert len(sub_mgr.get_all_filters(server_id)) == 1
            assert len(sub_mgr.get_all_subscriptions(server_id)) == 1

            assert self.get_object_count(sub_mgr, server_id) == 3

            sub_paths = [sub.path for sub in
                         sub_mgr.get_all_subscriptions(server_id)]
            sub_mgr.remove_subscriptions(server_id, sub_paths)

            dest_paths = [dest.path for dest in
                          sub_mgr.get_all_destinations(server_id)]
            sub_mgr.remove_destinations(
                server_id, dest_paths)

            for filter_ in sub_mgr.get_all_filters(server_id):
                sub_mgr.remove_filter(server_id, filter_.path)
            assert self.get_object_count(sub_mgr, server_id) == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author: Andreas Maier
#
# Tool that compares the symbols in the external API of two versions of a
# Python package.
# Invoke with --help for usage.

import sys
import argparse
import os
import fnmatch
import importlib
import re
import types
from pprint import pprint

ARGS = None
MYNAME=sys.argv[0]

def main():
    global ARGS

    parser = argparse.ArgumentParser(
        prog=MYNAME,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        usage="%s [-h|--help] [options] olddir newdir" % MYNAME,
        description=
"""
Compare the symbols in the external API of two versions of a Python package,
and print a report to stdout.

Package directories can be specified with relative or absolute path, and must
specify the package directory (inside of which there is the __init__.py file
for the package).
""",
        epilog=
"""
example:
  %s ../v1.0/foo foo
    Compare the foo package version in ../v1.0/foo with the version in ./foo.
""" % MYNAME)

    parser.add_argument("olddir", type=str,
                        help="path name of old package directory")
    parser.add_argument("newdir", type=str,
                        help="path name of new package directory")
    parser.add_argument("-f", "--full", dest='full', action='store_true',
                        help="list symbols of added or removed modules")
    parser.add_argument("-d", "--debug", dest='debug', action='store_true',
                        help="debug mode")
    parser.set_defaults(debug=False)
    parser.add_argument("-o", "--others", dest='others', action='store_true',
                        help="report symbol changes also when imported from other packages")
    parser.set_defaults(others=False)
    parser.add_argument("-i", "--imported", dest='imported', action='store_true',
                        help="report symbol changes also when imported from other modules of this package")
    parser.set_defaults(imported=False)
    parser.add_argument("-e", "--exported", dest='exported', action='store_true',
                        help="report symbol changes also for exported symbols (in __all__)")
    parser.set_defaults(exported=False)
    ARGS = parser.parse_args()

    rc = compare_package_dirs(ARGS.olddir, ARGS.newdir)
    return rc

def compare_package_dirs(old_pkgdir, new_pkgdir):

    old_pkgname = abs_modulename(old_pkgdir, old_pkgdir)
    new_pkgname = abs_modulename(new_pkgdir, new_pkgdir)
    if old_pkgname != new_pkgname:
        print("Error: Package names are not the same: %s / %s" % (old_pkgname, new_pkgname))
        return 1
    pkgname = old_pkgname

    print("Comparing public symbols of Python package: %s" % pkgname)
    print("    Old package directory: %s" % old_pkgdir)
    print("    New package directory: %s" % new_pkgdir)
    print("")
    print("This report covers the following kinds of symbols:")
    print("    Non-public symbols:                                                      %s" % False)
    print("    Symbols of added or removed modules (-f option):                         %s" % bool(ARGS.full))
    print("    Symbols imported from other packages (-o option):                        %s" % bool(ARGS.others))
    print("    Symbols imported from other modules of this package (-i option):         %s" % bool(ARGS.imported))
    print("    Symbols exported from this package (in __all__) (-e option):             %s" % bool(ARGS.exported))

    if ARGS.debug:
        print("Debug: Gathering information about modules in old package")
    old_modinfos = get_modinfos(old_pkgdir)
    if ARGS.debug:
        print("Debug: Gathering information about modules in new package")
    new_modinfos = get_modinfos(new_pkgdir)

    old_modnames = old_modinfos.keys()
    new_modnames = new_modinfos.keys()

    #print("Debug: Old module infos:")
    #pprint(old_modinfos)
    #print("Debug: New module infos:")
    #pprint(new_modinfos)

    added_modnames = set(new_modnames) - set(old_modnames)
    removed_modnames = set(old_modnames) - set(new_modnames)
    same_modnames = set(new_modnames).intersection(set(old_modnames))

    print("\nModule changes:")

    for mn in sorted(added_modnames):
        print("    Added module: %s" % mn)
        if ARGS.full:
            modinfo = new_modinfos[mn]
            syminfos = modinfo['syminfos']
            for sn in sorted(syminfos):
                si = syminfos[sn]
                print("        Symbol: {name} ({typename}, {modname})".format(**si))

    for mn in sorted(removed_modnames):
        print("    Removed module: %s" % mn)
        if ARGS.full:
            modinfo = old_modinfos[mn]
            syminfos = modinfo['syminfos']
            for sn in sorted(syminfos):
                si = syminfos[sn]
                print("        Symbol: {name} ({typename}, {modname})".format(**si))

    for mn in sorted(same_modnames):
        compare_modules(old_modinfos[mn], new_modinfos[mn], pkgname)

    return 0

def compare_modules(old_modinfo, new_modinfo, pkgname):

    mod_name = old_modinfo['name']

    old_syminfos = {}
    for sym in old_modinfo['syminfos']:
        si = old_modinfo['syminfos'][sym]
        sym_modname = si['modname']
        sym_pkgname = None if sym_modname is None else sym_modname.split('.')[0]
        exclude_reason = None
        if exclude(si):
            exclude_reason = EXCLUDE_REASON
        elif sym_pkgname is not None and sym_pkgname != pkgname and not ARGS.others:
            # None means unknown package origin, so we are not sure and don't exclude it
            exclude_reason = "imported from other package %s" % sym_pkgname
        elif sym_modname is not None and sym_modname != mod_name and not ARGS.imported:
            # None means unknown package origin, so we are not sure and don't exclude it
            exclude_reason = "imported from other module %s of this package" % sym_modname
        if not exclude_reason:
            old_syminfos[sym] = si
        else:
            if ARGS.debug:
                print("Debug: Excluding symbol %s defined in old module %s because: %s" % \
                      (sym, mod_name, exclude_reason))

    new_syminfos = {}
    for sym in new_modinfo['syminfos']:
        si = new_modinfo['syminfos'][sym]
        sym_modname = si['modname']
        sym_pkgname = None if sym_modname is None else sym_modname.split('.')[0]
        exclude_reason = None
        if exclude(si):
            exclude_reason = EXCLUDE_REASON
        elif sym_pkgname is not None and sym_pkgname != pkgname and not ARGS.others:
            # None means unknown package origin, so we are not sure and don't exclude it
            exclude_reason = "imported from other package %s" % sym_pkgname
        elif sym_modname is not None and sym_modname != mod_name and not ARGS.imported:
            # None means unknown package origin, so we are not sure and don't exclude it
            exclude_reason = "imported from other module %s of this package" % sym_modname
        if not exclude_reason:
            new_syminfos[sym] = si
        else:
            if ARGS.debug:
                print("Debug: Excluding symbol %s defined in new module %s because: %s" % \
                      (sym, mod_name, exclude_reason))

    added_symbols = set(new_syminfos.keys()) - set(old_syminfos.keys())
    removed_symbols = set(old_syminfos.keys()) - set(new_syminfos.keys())

    old_exp_syminfos = {}
    for sym in old_syminfos:
        if old_syminfos[sym]['exported']:
            old_exp_syminfos[sym] = old_syminfos[sym]

    new_exp_syminfos = {}
    for sym in new_syminfos:
        if new_syminfos[sym]['exported']:
            new_exp_syminfos[sym] = new_syminfos[sym]

    added_exp_symbols = set(new_exp_syminfos.keys()) - set(old_exp_syminfos.keys())
    removed_exp_symbols = set(old_exp_syminfos.keys()) - set(new_exp_syminfos.keys())

    have_report = len(added_symbols) + len(removed_symbols) > 0
    if ARGS.exported and not have_report:
        have_report = len(added_exp_symbols) + len(removed_exp_symbols) > 0
    if have_report:
        print("\nDifferences for module: %s" % mod_name)

    def where(sym_modname, our_modname):
        if sym_modname == our_modname:
            return 'defined here'
        if sym_modname is None:
            return 'unknown origin'
        return 'imported from %s' % sym_modname

    for sym in sorted(added_symbols):
        si = new_syminfos[sym]
        print("    Added   symbol: {0:30} ({1}, {2})".format(sym, si['typename'], where(si['modname'], mod_name)))

    for sym in sorted(removed_symbols):
        si = old_syminfos[sym]
        print("    Removed symbol: {0:30} ({1}, {2})".format(sym, si['typename'], where(si['modname'], mod_name)))

    if ARGS.exported:
        for sym in sorted(added_exp_symbols):
            si = new_syminfos[sym]
            print("    Added   exported symbol: {0:30} ({1}, {2})".format(sym, si['typename'], where(si['modname'], mod_name)))

        for sym in sorted(removed_exp_symbols):
            si = old_syminfos[sym]
            print("    Removed exported symbol: {0:30} ({1}, {2})".format(sym, si['typename'], where(si['modname'], mod_name)))

EXCLUDE_REASON = "non-public symbol"
def exclude(syminfo):

    sym = syminfo['name']
    if re.match(r'^__.+__$', sym): # special symbols
        return False
    if re.match(r'^__?.*', sym): # non-public symbols
        return True
    return False

def get_modinfos(pkgdir):

    mod_files = rglob(pkgdir, "*.py")
    pkgpath = os.path.abspath(pkgdir)
    pkgparentpath = os.path.dirname(pkgpath)

    pkgname = abs_modulename(pkgdir, pkgdir)
    unload_modules(pkgname)

    if pkgpath not in sys.path:
        if ARGS.debug:
            print("Debug: Inserting package dir into module search path: %s" % pkgpath)
        sys.path.insert(0, pkgpath)
    else:
        if ARGS.debug:
            print("Debug: Package dir is already in module search path: %s" % pkgpath)
        pass
    if pkgparentpath not in sys.path:
        if ARGS.debug:
            print("Debug: Inserting parent package dir into module search path: %s" % pkgparentpath)
        sys.path.insert(0, pkgparentpath)
    else:
        if ARGS.debug:
            print("Debug: Parent package dir is already in module search path: %s" % pkgparentpath)
        pass

    use_exec = True

    modinfos = {}

    for mod_file in mod_files:

        if ARGS.debug:
            print("Debug: Processing file: %s" % mod_file)

        rel_modname = rel_modulename(mod_file, pkgdir)
        abs_modname = abs_modulename(mod_file, pkgdir)

        if use_exec:

            module_import(abs_modname)
            mod_obj = sys.modules[abs_modname]
            mi = get_modinfos_from_mod_obj(mod_obj, abs_modname, mod_file)
            #if ARGS.debug:
            #    print("Debug: Got module infos:")
            #    pprint(mi)
            modinfos.update(mi)

            ## Not needed, because the normal import already determines
            ## wildcard information.
            #wc_sym_dict = wildcard_import(abs_modname)
            #if ARGS.debug:
            #    print("Debug: Got symbols: %r" % sorted(wc_sym_dict.keys()))
            #wc_mi = get_modinfos_from_sym_dict(wc_sym_dict, abs_modname, mod_file)
            #if ARGS.debug:
            #    print("Debug: Got module infos:")
            #    pprint(wc_mi)
            #modinfos.update(wc_mi)

        else:
            try:
                if rel_modname == '.':
                    if ARGS.debug:
                        print("Debug: Importing absolute module %s" % abs_modname)
                    mod_obj = importlib.import_module(abs_modname)
                else:
                    if ARGS.debug:
                        print("Debug: Importing relative module %s within package %s" % (rel_modname, pkgname))
                    mod_obj = importlib.import_module(rel_modname, pkgname)
            except ImportError as exc:
                if ARGS.debug:
                    print("Debug: Import failed (retrying): %s" % str(exc))
                try:
                    if ARGS.debug:
                        print("Debug: Importing absolute module %s" % abs_modname)
                    mod_obj = importlib.import_module(abs_modname)
                except ImportError as exc:
                    print("Error: Import failed (aborting): %s" % str(exc))
                    print("Loaded modules in package namespace:")
                    pprint([mod for mod in sorted(sys.modules.keys())
                            if mod == pkgname or mod.startswith(pkgname+'.')
                           ])
                    print("Module search path:")
                    pprint(sys.path)
                    raise

            assert abs_modname in sys.modules, "module %s is not in sys.modules after import" % abs_modname

            modinfos = get_modinfos_from_mod_obj(mod_obj, abs_modname, mod_file)

            del mod_obj # unload by garbage collection
            #del sys.modules[abs_modname] # this goes too far, probably

    if ARGS.debug:
        print("Debug: Removing package dir from module search path: %s" % pkgpath)
    sys.path.remove(pkgpath)
    if ARGS.debug:
        print("Debug: Removing parent package dir from module search path: %s" % pkgparentpath)
    sys.path.remove(pkgparentpath)

    return modinfos

def get_modinfos_from_mod_obj(mod_obj, abs_modname, mod_file):
    if not isinstance(mod_obj, types.ModuleType):
        raise TypeError("mod_obj parameter must be a module object.")
    sym_dict = mod_obj.__dict__
    return get_modinfos_from_sym_dict(sym_dict, abs_modname, mod_file)

def get_modinfos_from_sym_dict(sym_dict, abs_modname, mod_file):

    if not isinstance(sym_dict, dict):
        raise TypeError("sym_dict parameter must be a dict object.")

    all_symbols = sym_dict.keys()
    exp_symbols = sym_dict.get('__all__', all_symbols)

    unexpected_symbols = set(exp_symbols) - set(all_symbols)
    assert len(unexpected_symbols) == 0, "module %s has unexpected exported symbols: %r" %  sorted(unexpected_symbols)

    modinfos = {}
    mi = {}
    mi['name'] = abs_modname
    mi['file'] = mod_file
    syminfos = {}
    for sym in all_symbols:
        si = {}
        si['name'] = sym
        sym_obj = sym_dict[sym]
        if isinstance(sym_obj, types.ClassType):
            tn = 'old-style class'
        elif not hasattr(sym_obj, '__module__'):
            # e.g. tuple
            tn = type(sym_obj).__name__
        elif sym_obj.__module__ == '__builtin__':
            tn = type(sym_obj).__name__
        elif isinstance(sym_obj, types.TypeType):
            tn = 'new-style class'
        else:
            tn = type(sym_obj).__name__
        si['typename'] = tn
        si['modname'] = getattr(sym_obj, '__module__', None)
        si['exported'] = bool(sym in exp_symbols)
        syminfos[sym] = si
    mi['syminfos'] = syminfos
    modinfos[abs_modname] = mi

    return modinfos

def module_import(_wci_modname):
    _wci_stmt = 'import %s' % _wci_modname
    if ARGS.debug:
        print("Debug: Importing module: %s" % _wci_modname)
    exec(_wci_stmt)
    imported = locals()
    del imported['_wci_modname']
    del imported['_wci_stmt']
    return imported
    
def wildcard_import(_wci_modname):
    _wci_stmt = 'from %s import *' % _wci_modname
    if ARGS.debug:
        print("Debug: Importing * from module: %s" % _wci_modname)
    exec(_wci_stmt)
    imported = locals()
    del imported['_wci_modname']
    del imported['_wci_stmt']
    return imported
    
def unload_modules(pkgname):
    removals = []
    for modname in sys.modules:
        if modname == pkgname or modname.startswith(pkgname+'.'):
            removals.append(modname)
    for modname in removals:
        if ARGS.debug:
            print("Debug: Removing module from loaded modules: %s" % modname)
        del sys.modules[modname]

def abs_modulename(mod_file, pkgdir):
    mod_name = os.path.splitext(os.path.relpath(mod_file, os.path.dirname(pkgdir)))[0].replace(os.sep, '.')
    mod_name_parts = mod_name.split('.')
    if mod_name_parts[-1] == '__init__':
        mod_name = '.'.join(mod_name_parts[0:-1])
    return mod_name

def rel_modulename(mod_file, pkgdir):
    mod_name = os.path.splitext(os.path.relpath(mod_file, pkgdir))[0].replace(os.sep, '.')
    mod_name_parts = mod_name.split('.')
    if mod_name_parts[-1] == '__init__':
        mod_name = '.'.join(mod_name_parts[0:-1])
    mod_name = '.' + mod_name
    return mod_name

def rglob(path='.', pattern='*'):
    matches = []
    for root, dirnames, filenames in os.walk(path):
        for filename in fnmatch.filter(filenames, pattern):
            matches.append(os.path.join(root, filename))
    return matches

if __name__ == '__main__':
    sys.exit(main())



#
# (C) Copyright 2003,2004 Hewlett-Packard Development Company, L.P.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Tim Potter <tpot@hp.com>
#

"""
This section defines constants for two areas:

* CIM status codes (the ``CIM_ERR_*`` symbols). They are for example stored in
  :exc:`~pywbem.CIMError` exceptions.
* Default CIM namespace :data:`~pywbem.cim_constants.DEFAULT_NAMESPACE`. It is
  used as a default namespace for a connection (see
  :class:`~pywbem.WBEMConnection`) when no namespace is provided for an
  operation.

Note: For tooling reasons, the constants are shown in the namespace
``pywbem.cim_constants``. However, they are also available in the ``pywbem``
namespace and should be used from there.
"""

# disable flake8 tests for no whitespace before : and line too long
# for this file because of special formatting in the file.
# flake8: noqa: E203,E501

# This module is meant to be safe for 'import *'.

from ._utils import _format

__all__ = [
    'CIM_ERR_FAILED',
    'CIM_ERR_ACCESS_DENIED',
    'CIM_ERR_INVALID_NAMESPACE',
    'CIM_ERR_INVALID_PARAMETER',
    'CIM_ERR_INVALID_CLASS',
    'CIM_ERR_NOT_FOUND',
    'CIM_ERR_NOT_SUPPORTED',
    'CIM_ERR_CLASS_HAS_CHILDREN',
    'CIM_ERR_CLASS_HAS_INSTANCES',
    'CIM_ERR_INVALID_SUPERCLASS',
    'CIM_ERR_ALREADY_EXISTS',
    'CIM_ERR_NO_SUCH_PROPERTY',
    'CIM_ERR_TYPE_MISMATCH',
    'CIM_ERR_QUERY_LANGUAGE_NOT_SUPPORTED',
    'CIM_ERR_INVALID_QUERY',
    'CIM_ERR_METHOD_NOT_AVAILABLE',
    'CIM_ERR_METHOD_NOT_FOUND',
    'CIM_ERR_NAMESPACE_NOT_EMPTY',
    'CIM_ERR_INVALID_ENUMERATION_CONTEXT',
    'CIM_ERR_INVALID_OPERATION_TIMEOUT',
    'CIM_ERR_PULL_HAS_BEEN_ABANDONED',
    'CIM_ERR_PULL_CANNOT_BE_ABANDONED',
    'CIM_ERR_FILTERED_ENUMERATION_NOT_SUPPORTED',
    'CIM_ERR_CONTINUATION_ON_ERROR_NOT_SUPPORTED',
    'CIM_ERR_SERVER_LIMITS_EXCEEDED',
    'CIM_ERR_SERVER_IS_SHUTTING_DOWN',
    'DEFAULT_NAMESPACE'
]

# CIM status codes

#: A general error occurred that is not covered by a more specific error code.
CIM_ERR_FAILED = 1

#: Access to a CIM resource is not available to the client.
CIM_ERR_ACCESS_DENIED = 2

#: The target namespace does not exist.
CIM_ERR_INVALID_NAMESPACE = 3

#: One or more parameter values passed to the method are not valid.
CIM_ERR_INVALID_PARAMETER = 4

#: The specified class does not exist.
CIM_ERR_INVALID_CLASS = 5

#: The requested object cannot be found. The operation can be unsupported on
#: behalf of the WBEM server in general or on behalf of an implementation of a
#: management profile.
CIM_ERR_NOT_FOUND = 6

#: The requested operation is not supported on behalf of the WBEM server, or on
#: behalf of a provided class. If the operation is supported for a provided
#: class but is not supported for particular instances of that class, then
#: CIM_ERR_FAILED shall be used.
CIM_ERR_NOT_SUPPORTED = 7

#: The operation cannot be invoked on this class because it has subclasses.
CIM_ERR_CLASS_HAS_CHILDREN = 8

#: The operation cannot be invoked on this class because one or more instances
#: of this class exist.
CIM_ERR_CLASS_HAS_INSTANCES = 9

#: The operation cannot be invoked because the specified superclass does not
#: exist.
CIM_ERR_INVALID_SUPERCLASS = 10

#: The operation cannot be invoked because an object already exists.
CIM_ERR_ALREADY_EXISTS = 11

#: The specified property does not exist.
CIM_ERR_NO_SUCH_PROPERTY = 12

#: The value supplied is not compatible with the type.
CIM_ERR_TYPE_MISMATCH = 13

#: The query language is not recognized or supported.
CIM_ERR_QUERY_LANGUAGE_NOT_SUPPORTED = 14

#: The query is not valid for the specified query language.
CIM_ERR_INVALID_QUERY = 15

#: The extrinsic method cannot be invoked.
CIM_ERR_METHOD_NOT_AVAILABLE = 16

#: The specified extrinsic method does not exist.
CIM_ERR_METHOD_NOT_FOUND = 17

# 18 and 19 existed once and had been removed again.

#: The specified namespace is not empty.
#: *New in pywbem 0.9.*
CIM_ERR_NAMESPACE_NOT_EMPTY = 20

#: The enumeration identified by the specified context cannot be found, is in
#: a closed state, does not exist, or is otherwise invalid.
#: *New in pywbem 0.9.*
CIM_ERR_INVALID_ENUMERATION_CONTEXT = 21

#: The specified operation timeout is not supported by the WBEM server.
#: *New in pywbem 0.9.*
CIM_ERR_INVALID_OPERATION_TIMEOUT = 22

#: The pull operation has been abandoned due to execution of a concurrent
#: CloseEnumeration operation on the same enumeration.
#: *New in pywbem 0.9.*
CIM_ERR_PULL_HAS_BEEN_ABANDONED = 23

#: The attempt to abandon a concurrent pull operation on the same enumeration
#: failed. The concurrent pull operation proceeds normally.
#: *New in pywbem 0.9.*
CIM_ERR_PULL_CANNOT_BE_ABANDONED = 24

#: Using a a filter query in pulled enumerations is not supported by the WBEM
#: server.
#: *New in pywbem 0.9.*
#pylint: disable=invalid-name
CIM_ERR_FILTERED_ENUMERATION_NOT_SUPPORTED = 25

#: The WBEM server does not support continuation on error.
#: *New in pywbem 0.9.*
#pylint: disable=invalid-name
CIM_ERR_CONTINUATION_ON_ERROR_NOT_SUPPORTED = 26

#: The WBEM server has failed the operation based upon exceeding server limits.
#: *New in pywbem 0.9.*
CIM_ERR_SERVER_LIMITS_EXCEEDED = 27

#: The WBEM server is shutting down and cannot process the operation.
#: *New in pywbem 0.9.*
CIM_ERR_SERVER_IS_SHUTTING_DOWN = 28

#pylint: disable=line-too-long
_STATUSCODE2NAME = {
    CIM_ERR_FAILED                 : 'CIM_ERR_FAILED',
    CIM_ERR_ACCESS_DENIED          : 'CIM_ERR_ACCESS_DENIED',
    CIM_ERR_INVALID_NAMESPACE      : 'CIM_ERR_INVALID_NAMESPACE',
    CIM_ERR_INVALID_PARAMETER      : 'CIM_ERR_INVALID_PARAMETER',
    CIM_ERR_INVALID_CLASS          : 'CIM_ERR_INVALID_CLASS',
    CIM_ERR_NOT_FOUND              : 'CIM_ERR_NOT_FOUND',
    CIM_ERR_NOT_SUPPORTED          : 'CIM_ERR_NOT_SUPPORTED',
    CIM_ERR_CLASS_HAS_CHILDREN     : 'CIM_ERR_CLASS_HAS_CHILDREN',
    CIM_ERR_CLASS_HAS_INSTANCES    : 'CIM_ERR_CLASS_HAS_INSTANCES',
    CIM_ERR_INVALID_SUPERCLASS     : 'CIM_ERR_INVALID_SUPERCLASS',
    CIM_ERR_ALREADY_EXISTS         : 'CIM_ERR_ALREADY_EXISTS',
    CIM_ERR_NO_SUCH_PROPERTY       : 'CIM_ERR_NO_SUCH_PROPERTY',
    CIM_ERR_TYPE_MISMATCH          : 'CIM_ERR_TYPE_MISMATCH',
    CIM_ERR_QUERY_LANGUAGE_NOT_SUPPORTED : 'CIM_ERR_QUERY_LANGUAGE_NOT_SUPPORTED',
    CIM_ERR_INVALID_QUERY          : 'CIM_ERR_INVALID_QUERY',
    CIM_ERR_METHOD_NOT_AVAILABLE   : 'CIM_ERR_METHOD_NOT_AVAILABLE',
    CIM_ERR_METHOD_NOT_FOUND       : 'CIM_ERR_METHOD_NOT_FOUND',
    CIM_ERR_NAMESPACE_NOT_EMPTY    : 'CIM_ERR_NAMESPACE_NOT_EMPTY',
    CIM_ERR_INVALID_ENUMERATION_CONTEXT    : 'CIM_ERR_INVALID_ENUMERATION_CONTEXT',
    CIM_ERR_INVALID_OPERATION_TIMEOUT      : 'CIM_ERR_INVALID_OPERATION_TIMEOUT',
    CIM_ERR_PULL_HAS_BEEN_ABANDONED        : 'CIM_ERR_PULL_HAS_BEEN_ABANDONED',
    CIM_ERR_PULL_CANNOT_BE_ABANDONED       : 'CIM_ERR_PULL_CANNOT_BE_ABANDONED',
    CIM_ERR_FILTERED_ENUMERATION_NOT_SUPPORTED : 'CIM_ERR_FILTERED_ENUMERATION_NOT_SUPPORTED',
    CIM_ERR_CONTINUATION_ON_ERROR_NOT_SUPPORTED : 'CIM_ERR_CONTINUATION_ON_ERROR_NOT_SUPPORTED',
    CIM_ERR_SERVER_LIMITS_EXCEEDED         : 'CIM_ERR_SERVER_LIMITS_EXCEEDED',
    CIM_ERR_SERVER_IS_SHUTTING_DOWN        : 'CIM_ERR_SERVER_IS_SHUTTING_DOWN',
}

def _statuscode2name(status_code):
    """Return the symbolic name for a CIM status code."""
    try:
        s = _STATUSCODE2NAME[status_code]
    except KeyError:
        s = _format("Invalid status code {0}", status_code)
    return s


_STATUSCODE2STRING = {
    CIM_ERR_FAILED                 : 'A general error occurred',
    CIM_ERR_ACCESS_DENIED          : 'Resource not available',
    CIM_ERR_INVALID_NAMESPACE      : 'The target namespace does not exist',
    CIM_ERR_INVALID_PARAMETER      : 'Parameter value(s) invalid',
    CIM_ERR_INVALID_CLASS          : 'The specified Class does not exist',
    CIM_ERR_NOT_FOUND              : 'Requested object could not be found',
    CIM_ERR_NOT_SUPPORTED          : 'Operation not supported',
    CIM_ERR_CLASS_HAS_CHILDREN     : 'Class has subclasses',
    CIM_ERR_CLASS_HAS_INSTANCES    : 'Class has instances',
    CIM_ERR_INVALID_SUPERCLASS     : 'Superclass does not exist',
    CIM_ERR_ALREADY_EXISTS         : 'Object already exists',
    CIM_ERR_NO_SUCH_PROPERTY       : 'Property does not exist',
    CIM_ERR_TYPE_MISMATCH          : 'Value incompatible with type',
    CIM_ERR_QUERY_LANGUAGE_NOT_SUPPORTED   : 'Query language not supported',
    CIM_ERR_INVALID_QUERY          : 'Query not valid',
    CIM_ERR_METHOD_NOT_AVAILABLE   : 'Extrinsic method not executed',
    CIM_ERR_METHOD_NOT_FOUND       : 'Extrinsic method does not exist',
    CIM_ERR_NAMESPACE_NOT_EMPTY    : 'Namespace not empty',
    CIM_ERR_INVALID_ENUMERATION_CONTEXT    : 'Enumeration context is invalid',
    CIM_ERR_INVALID_OPERATION_TIMEOUT      : 'Operation timeout not supported',
    CIM_ERR_PULL_HAS_BEEN_ABANDONED        : 'Pull operation has been abandoned',
    CIM_ERR_PULL_CANNOT_BE_ABANDONED       : 'Attempt to abandon a pull operation failed',
    CIM_ERR_FILTERED_ENUMERATION_NOT_SUPPORTED : 'Filtered pulled enumeration not supported',
    CIM_ERR_CONTINUATION_ON_ERROR_NOT_SUPPORTED : 'WBEM server does not support continuation on error',
    CIM_ERR_SERVER_LIMITS_EXCEEDED         : 'WBEM server limits exceeded',
    CIM_ERR_SERVER_IS_SHUTTING_DOWN        : 'WBEM server is shutting down',
}

def _statuscode2string(status_code):
    """Return a short message for a CIM status code."""
    try:
        s = _STATUSCODE2STRING[status_code]
    except KeyError:
        s = _format("Invalid status code {0}", status_code)
    return s


# Provider types
PROVIDERTYPE_CLASS = 1
PROVIDERTYPE_INSTANCE = 2
PROVIDERTYPE_ASSOCIATION = 3
PROVIDERTYPE_INDICATION = 4
PROVIDERTYPE_METHOD = 5
PROVIDERTYPE_CONSUMER = 6               # Indication consumer
PROVIDERTYPE_QUERY = 7

DEFAULT_NAMESPACE = 'root/cimv2'        #: Default CIM namespace
